import numpy as np
import os.path
from itertools import *
import pandas as pd
import logging
import time
import pysnptools.util as pstutil
import numbers

try:
    from builtins import int
except Exception:
    pass


class PstReader(object):
    """A PstReader is one of three things:

    * A :class:`.PstData` class that holds matrix data in memory:

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> type(data1.val).__name__ # The val property is an ndarray of values
        'ndarray'
        >>> data1.row_count #prints the number of rows in this in-memory data
        3

    * A class such as :class:`.PstNpz` for you to specify data in file. For example,

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> pstnpz_file = example_file('tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk = PstNpz(pstnpz_file)
        >>> print(on_disk) # prints the name of the file reader
        PstNpz(...tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk.col_count # prints the number of columns (but doesn't read any matrix values)
        1015

    * A subset of any PstReader, specified with "[ *row_index* , *col_index* ]", to read just some values.

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> pstnpz_file = example_file('tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk = PstNpz(pstnpz_file)
        >>> subset_on_disk = on_disk[[3,4],::2] # specification for a subset of the data on disk. No values are read yet.
        >>> print(subset_on_disk.col_count) # prints the number of columns in this subset (but still doesn't read any values)
        508
        >>> print(subset_on_disk) #prints a specification of 'subset_on_disk'
        PstNpz(...tests/datasets/all_chr.maf0.001.N300.pst.npz')[[3,4],::2]
        >>> data_subset = subset_on_disk.read() # efficiently (if practical) reads the specified subset of values from the disk
        >>> print(data_subset) # prints the specification of the in-memory information
        PstData(PstNpz(...tests/datasets/all_chr.maf0.001.N300.pst.npz')[[3,4],::2])
        >>> int(data_subset.val.shape[0]),int(data_subset.val.shape[1]) # The dimensions of the ndarray of values
        (2, 508)

    The PstReaders Classes

        ================================== ================================== ====================== =====================
        *Class*                            *Format*                           *Random Access*        *Suffixes*
        :class:`.PstData`                  in-memory floats                   Yes                    *n/a*
        :class:`.PstNpz`                   binary, floats                     No                     .pst.npz,.snp.npz,
                                                                                                     .kernel.npz
        :class:`.PstHdf5`                  binary, floats                     Yes                    .pst.hdf5,.snp.hdf5,
                                                                                                     .kernel.hdf5
        :class:`.PstMemMap`                mem-mapped floats                  Yes                    .pst.memmap,.snp.memmap
        various :class:`.SnpReader`        varies                             varies                 varies
        various :class:`.KernelReader`     varies                             varies                 varies
        ================================== ================================== ====================== =====================


            A :class:`.SnpReader` and :class:`.KernelReader` are each a kind of :class:`.PstReader`. They have some restrictions summarized here:
            ================================== =============== ============ ============ ==================== ====================
            *Class*                            *val type*      *row type*   *col type*   *row_property type*  *col_property type*
            :class:`.PstReader`                float           any          any          any                  any
            :class:`.SnpReader`                float           str,str      str          none                 float,float,float
            :class:`.KernelReader`             float           str,str      str,str      none                 none
            ================================== =============== ============ ============ ==================== ====================

            For convenience, they allow additional ways to access rows and columns.

            ================================== =============== ============ ============ ==================== ====================
            *Class*                            *val name*      *row name*   *col name*   *row_property name*  *col_property name*
            :class:`.PstReader`                val             row          col          row_property         col_property
            :class:`.SnpReader`                val             iid,row      sid,col      none                 col_property,pos
            :class:`.KernelReader`             val             iid0,row,iid iid1,col,iid none                 none
            ================================== =============== ============ ============ ==================== ====================

            :Note: A :attr:`KernelReader.iid` may be used when :attr:`KernelReader.iid0` is equal to :attr:`KernelReader.iid1`

    Methods & Properties:

        Every PstReader, such as :class:`.PstNpz` and :class:`.PstData`, has these properties: :attr:`row`, :attr:`row_count`, :attr:`col`, :attr:`col_count`,
        :attr:`row_property`, :attr:`col_property` and these methods: :meth:`read`, :meth:`row_to_index`, :meth:`col_to_index`. See below for details.

        :class:`.PstData` is a PstReader so it supports the above properties and methods. In addition, it supports property :attr:`PstData.val` and equality testing.
        See below for details.

    Rows and Cols:

        Example:

        >>> from pysnptools.pstreader import PstHdf5
        >>> from pysnptools.util import print2 #print bytes strings and Unicode strings the same
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file('pysnptools/examples/toydata.iidmajor.snp.hdf5')
        >>> on_disk = PstHdf5(hdf5_file) # PstHdf5 can load .pst.hdf5, .snp.hdf5, and kernel.hdf5
        >>> print2(on_disk.row[:3]) # print the first three rows
        [['per0' 'per0']
         ['per1' 'per1']
         ['per2' 'per2']]
        >>> print2(on_disk.col[:7]) # print the first seven columns
        ['null_0' 'null_1' 'null_2' 'null_3' 'null_4' 'null_5' 'null_6']
        >>> print2(on_disk.row_to_index([[b'per2', b'per2'],[b'per1', b'per1']])) #Find the indexes for two rows.
        [2 1]

    When Matrix Data is Read:

        Matrix data can be enormous so we generally avoid reading it to the degree practical. Specifically,

        * Constructing and printing a PstReader causes no file reading. For example, these commands read no data:

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> print(on_disk) # Print the PstHdf5 PstReader specification. No data is read.
            PstHdf5(...pysnptools/examples/toydata.iidmajor.snp.hdf5')
            >>> subset_on_disk = on_disk[[3,4],::2] # Construct a subsetting PstReader. No data is read.
            >>> print(subset_on_disk) # print the subset PstReader. No data is read.
            PstHdf5(...pysnptools/examples/toydata.iidmajor.snp.hdf5')[[3,4],::2]

        * Properties and methods related to the rows and columns (to the degree practical) read only row and col data from the disk,
          not value data. Moreover, the row and col data is read from file only once. Consider these commands:

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> print2(on_disk.col[:7]) # without reading any values data from disk, read the row and col data from disk, cache it, and then print the first seven cols.
            ['null_0' 'null_1' 'null_2' 'null_3' 'null_4' 'null_5' 'null_6']
            >>> print(on_disk.col_to_index([b'null_7',b'null_2'])) #use the cached col information to find the indexes of 'null_7' and 'null_2'. (No data is read from disk.)
            [7 2]

        * The only method that reads values from file (to the degree practical) is :meth:`read`. For example:

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> data1 = on_disk.read() #read all the values from disk, creating a new PstData instance that keeps these values in memory
            >>> print(data1.val[0,2]) # print the value for the row with index 0 and the col with index 2. (No data is read from disk.)
            2.0

        * If you request the values for only a subset of the rows or columns, (to the degree practical) only that subset will be read from disk.
          for example:

            >>> hdf5_file2 = example_file('tests/datasets/all_chr.maf0.001.N300.hdf5')
            >>> on_disk = PstHdf5(hdf5_file2)[[3,4],::2] # Construct a subsetting PstReader. No data is read.
            >>> data_subset = subset_on_disk.read() # from disk, read the values for the rows with index 3 and 4 AND cols with even numbered indexes.
            >>> print(data_subset.val[0,2]) # print the value with subset row index 0 and col index 2 (corresponding to row index 3 and col index 4 in the full data). No data is read from disk.
            1.0

    When Matrix Data is Re-Read and Copied:

        Every time you call a PstReader's :meth:`read` method, the PstReader re-reads the value data and returns a new in-memory :class:`.PstData`
        (with :attr:`PstData.val` property containing a new ndarray of the values).

        Here is an example of what not to do, because it causes all the matrix value data to be read twice.

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> # Not recommended because it inefficiently reads all the values twice.
            >>> print(on_disk.read().val[0,2]) # read all values into a new PstData, print a value
            2.0
            >>> print(on_disk.read().val[0,3]) # read all values (again) into a second new PstData, print a value
            2.0

        Here are two efficient alternatives. First, if all values can all fit in memory, read them once into a :class:`PstData` and then
        access that :class:`PstData` multiple times.

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> data1 = on_disk.read() # read all values into a new PstData
            >>> print(data1.val[0,2]) # print a value from data1's in-memory ndarray
            2.0
            >>> print(data1.val[0,3]) # print another value from data1's in-memory ndarray.
            2.0

        Second, if the value data is too large to fit in memory, use subsetting to read only the values of interest from disk.

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> print(on_disk[0,2].read().val[0,0]) #Define the subset of data and read only that subset from disk.
            2.0
            >>> print(on_disk[0,3].read().val[0,0]) #Define a second subset of data and read only that subset from disk.
            2.0

        Because the in-memory :class:`.PstData` class is a kind of PstReader, you may read from it, too.
        Doing so create a new :class:`.PstData` instance containing a copy of the matrix values in a new ndarray.

            >>> on_disk = PstHdf5(hdf5_file) # Construct a PstHdf5 PstReader. No data is read.
            >>> data1 = on_disk.read() # read all matrix values from disk into a new PstData
            >>> print(data1.val is data1.val) # Do the in-memory SNP values use the same memory as themselves? Yes
            True
            >>> data2 = data1.read() # copy all the matrix values into a new ndarray in a new PstData
            >>> print(data2.val is data1.val) # Do the two ndarrays of in-memory matrix values use the same memory?
            False


    Avoiding Unwanted ndarray Allocations

        You may want a subset of matrix values from an in-memory :class:`PstData` and you may know that this subset and the original :class:`PstData`
        can safely share the memory of the ndarray of matrix values. For this case, the :meth:`read` has optional parameters called view_ok and order. If you override
        the defaults of "view_ok=False,order='F'" with "view_ok=True,order='A', the :meth:`read` will, if practical, return a new
        :class:`PstData` with a ndarray shares memory with the original ndarray.
        Use these parameters with care because any change to either ndarray will effect
        the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
        share memory and so it may ignore your suggestion and allocate a new ndarray anyway.

            >>> on_disk = PstNpz(pstnpz_file) # Construct a PstNpz PstReader. No data is read.
            >>> data1 = on_disk.read() # read all data from disk into a PstData with a new ndarray
            >>> column01 = data1[:,0:1].read(view_ok=True,order='A') #create PstData with the data from just the first two columns. Sharing memory is OK. The memory may be laid out in any order (that is col-major and row-major are both OK).
            >>> import numpy as np
            >>> #print(np.may_share_memory(data1.val, column01.val)) # Do the two ndarray's share memory? They could (but currently they won't)
            >>> column201 = data1[:,[2,0,1]].read(view_ok=True,order='A') #create PstData with the data from three columns, permuted. Sharing memory is OK.
            >>> print(np.may_share_memory(data1.val, column201.val)) # Do the two ndarray's share memory? No, ndarray decided that this indexing was too complex for sharing.
            False

    Creating Subsetting PstReaders with Indexing

        You often don't want to read the matrix values for all rows and cols. You can use indexing to create a subsetting PstReader that
        will read only the matrix values of interest.

        PstReaders support the indexing formats supported by ndarray plus two generalizations. Here are examples of indexing with an array
        of indexes, with slicing, and with an array of Booleans.

            >>> on_disk = PstNpz(pstnpz_file) # Specify some data on disk in PstNpz format
            >>> subset_reader_1 = on_disk[[3,-1],:] #index with an array of indexes (negatives count from end)
            >>> print((subset_reader_1.row_count, subset_reader_1.col_count))
            (2, 1015)
            >>> data1 = subset_reader_1.read() # read just the two rows of interest from the disk
            >>> subset_reader_2 = on_disk[:,:0:-2] #index with a slice
            >>> print((subset_reader_2.row_count, subset_reader_2.col_count))
            (300, 507)
            >>> boolindexes = [s.startswith(b'23_') for s in on_disk.col] # create a Boolean index of cols that start '23_'
            >>> subset_reader_3 = on_disk[:,boolindexes] #index with array of Booleans
            >>> print((subset_reader_3.row_count, subset_reader_3.col_count))
            (300, 24)

        The first generalization over what ndarray offers is full indexing on both the row dimension and the col dimension, in other words,
        full multidimensional indexing. For example,

            >>> on_disk = PstNpz(pstnpz_file) # Specify some data on disk in PstNpz format
            >>> subset_reader_4 = on_disk[[3,4],:0:-2] # index on two dimensions at once
            >>> print((subset_reader_4.row_count, subset_reader_4.col_count))
            (2, 507)

        The second generalization is indexing on a single integer index.

            >>> on_disk = PstNpz(pstnpz_file) # Specify some data on disk in PstNpz format
            >>> subset_reader_5 = on_disk[5,:] #index with single integer
            >>> print((subset_reader_5.row_count, subset_reader_5.col_count))
            (1, 1015)

        Indexing is also useful when you have matrix values in memory via a :class:`PstData` index and want to copy a subset of those values.
        While you could instead index directly on the :attr:`PstData.val` ndarray, by indexing on the :class:`PstData` instance you
        also get row and col information.

            >>> on_disk = PstNpz(pstnpz_file) # Specify some data on disk in PstNpz format
            >>> data1 = on_disk.read() # read all matrix values into memory
            >>> print2(data1.col[:9]) # print the first 9 cols
            ['1_12' '1_34' '1_10' '1_35' '1_28' '1_25' '1_36' '1_39' '1_4']

            >>> data_subset = data1[:,::2].read(view_ok=True,order='A') # create a copy or view with every other col
            >>> print2(data_subset.col[:9]) # print the first 9 cols in the subset
            ['1_12' '1_10' '1_28' '1_36' '1_4' '1_11' '1_32' '1_9' '1_17']


        You can apply indexing on top of indexing to specify subsets of subsets of data to read. In this example,
        only the column values for every 16th col is actually read from the disk.

            >>> # These are just PstReaders, nothing is read from disk yet
            >>> on_disk = PstNpz(pstnpz_file) # Specify some data on disk in PstNpz format
            >>> half_reader = on_disk[:,::2] # a reader for half the cols
            >>> quarter_reader = half_reader[:,::2] # a reader for half of half the cols
            >>> sixteenth_reader = quarter_reader[:,::2][:,::2] # a reader for half of half of half of half the cols
            >>> print(sixteenth_reader) #Print the specification of this reader
            PstNpz(...tests/datasets/all_chr.maf0.001.N300.pst.npz')[:,::2][:,::2][:,::2][:,::2]
            >>> # Now we read from disk. Only values for one col in every 16 will be read.
            >>> data_sixteenth = sixteenth_reader.read()
            >>> print(data_sixteenth.val[0,3])
            2.0

    The :meth:`read` Method

        By default the :meth:`read` returns a ndarray of numpy.float64 laid out in memory in F-contiguous order (row-index varies the fastest). You may, instead,
        ask for numpy.float32 or for C-contiguous order or any order. See :meth:`read` for details.

    Details of Methods & Properties:
    """

    def __init__(
        self, *args, **kwargs
    ):  # Ignore any inputs because our parent is 'object'
        super(PstReader, self).__init__()

    @property
    def row(self):
        """A ndarray of the row ids. Each id can be anything, for example, a string, an array of two strings, a number, etc.

        :rtype: ndarray (length :attr:`.row_count`)

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data.
        Moreover, the row and col data is read from file only once.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> print(data1.row[:2]) # print the first two row ids
        ['a' 'b']
        >>> data2 = PstData(row=[[1,0],[2,0],[2,1]],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> print(data2.row[:2]) # print the first two row ids
        [[1 0]
         [2 0]]
        """
        raise NotImplementedError

    @property
    def row_count(self):
        """number of rows

        :rtype: integer

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.
        """
        return len(self.row)

    @property
    def col(self):
        """A ndarray of the cols id. Each id can be anything, for example, a string, an array of two strings, a number, etc.

        :rtype: ndarray (length :attr:`.col_count`)

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data.
        Moreover, the row and col data is read from file only once.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> print(data1.col[-1]) # print the last column id.
        z
        >>> data2 = PstData(row=['a','b','c'],col=[[1,0],[2,0]],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> print(data2.col[-1]) # print the last column id.
        [2 0]
        """
        raise NotImplementedError

    @property
    def col_count(self):
        """number of cols

        :rtype: integer

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        """
        return len(self.col)

    @property
    def shape(self):
        """number of rows and number of cols

        :rtype: tuple of two integers

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        """
        return (len(self.row), len(self.col))

    @property
    def row_property(self):
        """A ndarray of the additional information for each row. Each element is a ndarray.

        :rtype: 1- or 2-dimensional ndarray (length :attr:`.row_count`)

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> print(data1.row_property[:2]) # print the first two row property values
        ['A' 'B']
        """
        raise NotImplementedError

    @property
    def col_property(self):
        """A ndarray of the additional information for each col. Each element is a ndarray.

        :rtype: 1- or 2-dimensional ndarray (length :attr:`.col_count`)

        This property (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        :Example:

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> pstnpz_file = example_file('tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk = PstNpz(pstnpz_file)
        >>> print(on_disk.col_property[:3]) # print column information for the first three cols: #The '...' is an optional space
        [[1.         0.00800801 0.        ]
         [1.         0.023023   1.        ]
         [1.         0.0700701  4.        ]]
        """
        raise NotImplementedError

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        raise NotImplementedError

    # !!check that views always return contiguous memory by default
    def read(
        self,
        order="F",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
    ):
        """Reads the matrix values and returns a :class:`.PstData` (with :attr:`PstData.val` property containing a new ndarray of the matrix values).

        :param order: {'F' (default), 'C', 'A'}, optional -- Specify the order of the ndarray. If order is 'F' (default),
            then the array will be in F-contiguous order (row-index varies the fastest).
            If order is 'C', then the returned array will be in C-contiguous order (col-index varies the fastest).
            If order is 'A', then the :attr:`PstData.val`
            ndarray may be in any order (either C-, Fortran-contiguous).
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`PstData.val` ndarray.
        :type dtype: data-type

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool


        :param view_ok: optional -- If False (default), allocates new memory for the :attr:`PstData.val`'s ndarray. If True,
            if practical and reading from a :class:`PstData`, will return a new
            :class:`PstData` with a ndarray shares memory with the original :class:`PstData`.
            Typically, you'll also wish to use "order='A'" to increase the chance that sharing will be possible.
            Use these parameters with care because any change to either ndarray will effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.
        :type view_ok: bool

        :param num_threads: optional -- The number of threads with which to read data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int


        :rtype: :class:`.PstData`

        Calling the method again causes the matrix values to be re-read and creates a new in-memory :class:`.PstData` with a new ndarray of matrix values.

        If you request the values for only a subset of the sids or iids, (to the degree practical) only that subset will be read from disk.

        :Example:

        >>> from pysnptools.pstreader import PstHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file('pysnptools/examples/toydata.iidmajor.snp.hdf5')
        >>> on_disk = PstHdf5(hdf5_file) # Specify matrix data on disk
        >>> pstdata1 = on_disk.read() # Read all the matrix data returning a PstData instance
        >>> print(type(pstdata1.val).__name__) # The PstData instance contains a ndarray of the data.
        ndarray
        >>> subset_pstdata = on_disk[:,::2].read() # From the disk, read matrix values for every other sid
        >>> print(subset_pstdata.val[0,0]) # Print the first matrix value in the subset
        1.0
        >>> subsub_pstdata = subset_pstdata[:10,:].read(order='A',view_ok=True) # Create an in-memory subset of the subset with matrix values for the first ten iids. Share memory if practical.
        >>> import numpy as np
        >>> # print(np.may_share_memory(subset_snpdata.val, subsub_snpdata.val)) # Do the two ndarray's share memory? They could. Currently they won't.
        """
        dtype = np.dtype(dtype)
        val = self._read(
            None, None, order, dtype, force_python_only, view_ok, num_threads
        )
        from pysnptools.pstreader import PstData

        ret = PstData(
            self.row,
            self.col,
            val,
            row_property=self.row_property,
            col_property=self.col_property,
            name=str(self),
        )
        return ret

    def row_to_index(self, list):
        """Takes a list of row ids and returns a list of index numbers

        :param list: list of rows
        :type order: list

        :rtype: ndarray of int

        This method (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        :Example:

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> pstnpz_file = example_file('tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk = PstNpz(pstnpz_file) # Specify matrix data on disk
        >>> print(on_disk.row_to_index([[b'POP1',b'44'],[b'POP1',b'12']])) #Find the indexes for two rows.
        [2 1]
        """
        if not hasattr(self, "_row_to_index"):
            self._row_to_index = {}
            for index, item in enumerate(self.row):
                key = self._makekey(item)
                if key in self._row_to_index:
                    raise Exception(
                        "Expect row to appear in data only once. ({0})".format(key)
                    )
                self._row_to_index[key] = index
        index = np.fromiter(
            (self._row_to_index[PstReader._makekey(item1)] for item1 in list), np.int_
        )
        return index

    def col_to_index(self, list):
        """Takes a list of column ds and returns a list of index numbers

        :param list: list of cols
        :type list: list

        :rtype: ndarray of int

        This method (to the degree practical) reads only row and col data from the disk, not matrix value data. Moreover, the row and col data is read from file only once.

        :Example:

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> pstnpz_file = example_file('tests/datasets/all_chr.maf0.001.N300.pst.npz')
        >>> on_disk = PstNpz(pstnpz_file) # Specify matrix data on disk
        >>> print(on_disk.col_to_index([b'1_10',b'1_13'])) #Find the indexes for two cols.
        [2 9]
        """
        if not hasattr(self, "_col_to_index"):
            logging.debug("Creating _col_to_index")
            col_set = None
            try:
                col_set = set(self.col)
            except Exception:
                pass

            if col_set is not None:
                assert (
                    len(col_set) == self.col_count
                ), "Expect col to appear in data only once."
                self._col_to_index = {
                    item: index for index, item in enumerate(self.col)
                }
            else:
                col_list = [PstReader._makekey(item) for item in self.col]
                assert (
                    len(set(col_list)) == self.col_count
                ), "Expect col to appear in data only once."
                self._col_to_index = {
                    item: index for index, item in enumerate(col_list)
                }
            logging.debug("Finished creating _col_to_index")
        index = np.fromiter(
            (self._col_to_index[PstReader._makekey(item1)] for item1 in list), np.int_
        )
        return index

    @staticmethod
    def _makekey(item):
        if isinstance(item, str):
            return item
        if isinstance(
            item, (numbers.Integral, float)
        ):  # return quickly from known items
            return item
        try:
            hash(item)
            return item
        except Exception:
            pass

        return tuple((PstReader._makekey(subitem) for subitem in item))

    def __getitem__(self, row_indexer_and_col_indexer):
        from pysnptools.pstreader._subset import _PstSubset

        row_indexer, col_indexer = row_indexer_and_col_indexer
        return _PstSubset(self, row_indexer, col_indexer)

    def copyinputs(self, copier):
        raise NotImplementedError

    @staticmethod
    def _is_all_slice(index_or_none):
        if index_or_none is None:
            return True
        return isinstance(index_or_none, slice) and index_or_none == slice(None)

    @staticmethod
    def _make_sparray_or_slice(indexer):
        if indexer is None:
            return slice(None)

        if isinstance(indexer, np.ndarray):
            return PstReader._process_ndarray(indexer)

        if isinstance(indexer, slice):
            return indexer

        if np.isscalar(indexer):
            assert isinstance(
                indexer, numbers.Integral
            ), "Expect scalar indexes to be integers"
            return np.array([indexer])

        return PstReader._process_ndarray(np.array(indexer))

    @staticmethod
    def _process_ndarray(indexer):
        if (
            len(indexer) == 0
        ):  # If it's zero length, the type is unreliable and unneeded.
            return np.zeros((0), dtype=np.integer)
        if indexer.dtype == bool:
            return np.arange(len(indexer), dtype=np.integer)[indexer]
        assert np.issubdtype(indexer.dtype, np.integer), "Indexer of unknown type"
        return indexer

    @staticmethod
    def _make_sparray_from_sparray_or_slice(count, indexer):
        if isinstance(indexer, slice):
            return np.arange(*indexer.indices(count), dtype="uintp")
        result = np.ascontiguousarray(
            (np.arange(count, dtype="uintp")[indexer]).reshape(-1), dtype="uintp"
        )
        return result

    @staticmethod
    def _array_properties_are_ok(val, order, dtype):
        dtype = np.dtype(dtype)

        if val.dtype != dtype:
            return False
        if order == "F":
            return val.flags["F_CONTIGUOUS"]
        elif order == "C":
            return val.flags["C_CONTIGUOUS"]

        return True

    def _apply_sparray_or_slice_to_val(
        self,
        val,
        row_indexer_or_none,
        col_indexer_or_none,
        order,
        dtype,
        force_python_only,
        num_threads,
    ):
        dtype = np.dtype(dtype)

        if (
            PstReader._is_all_slice(row_indexer_or_none)
            and PstReader._is_all_slice(col_indexer_or_none)
            and (
                order == "A"
                or (order == "F" and val.flags["F_CONTIGUOUS"])
                or (order == "C" and val.flags["C_CONTIGUOUS"])
            )
            and (dtype is None or val.dtype == dtype)
        ):
            return val, True

        row_indexer = PstReader._make_sparray_or_slice(row_indexer_or_none)
        col_indexer = PstReader._make_sparray_or_slice(col_indexer_or_none)
        if not force_python_only:
            row_index = PstReader._make_sparray_from_sparray_or_slice(
                self.row_count, row_indexer
            )
            col_index = PstReader._make_sparray_from_sparray_or_slice(
                self.col_count, col_indexer
            )
            sub_val = pstutil.sub_matrix(
                val,
                row_index,
                col_index,
                order=order,
                dtype=dtype,
                num_threads=num_threads,
            )
            return sub_val, False

        if PstReader._is_all_slice(row_indexer) or PstReader._is_all_slice(col_indexer):
            sub_val = val[row_indexer, col_indexer]  # !!is this faster than the C++?
        else:
            row_index = PstReader._make_sparray_from_sparray_or_slice(
                self.row_count, row_indexer
            )
            col_index = PstReader._make_sparray_from_sparray_or_slice(
                self.col_count, col_indexer
            )
            # See http://stackoverflow.com/questions/21349133/numpy-array-integer-indexing-in-more-than-one-dimension
            sub_val = val[row_index.reshape(-1, 1), col_index]

        assert len(sub_val.shape) in {
            2,
            3,
        }, "Expect result of subsetting to be 2 or 3 dimensional"

        if not PstReader._array_properties_are_ok(sub_val, order, dtype):
            if order is None:
                order = "K"
            if dtype is None:
                dtype = sub_val.dtype
            sub_val = sub_val.astype(dtype, order, copy=True)

        shares_memory = np.may_share_memory(val, sub_val)
        assert PstReader._array_properties_are_ok(sub_val, order, dtype)
        return sub_val, shares_memory


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test
    print("done")



import numpy as np
import scipy as sp
import unittest
import logging
import os
import doctest

# from pysnptools.snpreader import SnpData, Bed, SnpNpz -- putting this here would cause a loop


def snp_gen(
    fst,
    dfr,
    iid_count,
    sid_count,
    maf_low=0.05,
    maf_high=0.5,
    seed=0,
    sibs_per_family=10,
    freq_pop_0=0.5,
    chr_count=None,
    label_with_pop=False,
):
    """Generates a random :class:`.SnpData`

    :param fst: Degree of Population Structure, e.g. 0 (a special case), 0.005, 0.01, 0.05, 0.1
    :type fst: float

    :param dfr: Degree of Family Relatedness, the fraction of individuals belonging to a family, ie. fracSibs, e.g. 0.0, 0.5, 0.6, 0.7, 0.8, 0.9
    :type dfr: float

    :param iid_count: The number of individuals to generate. Because of rounding the actual number may be less.
    :type iid_count: int

    :param sid_count: The number of snps to generate.
    :type sid_count: int

    :param maf_low: (default .05) lower bound of uniformly-generated Minor allele frequency
    :type maf_low: float

    :param maf_high: (default .5) upper bound of uniformly-generated Minor allele frequency
    :type maf_high: float

    :param seed: (default 0) Random seed
    :type seed: int

    :param sibs_per_family: (default 10) number of siblings in each family
    :type sibs_per_family: int

    :param freq_pop_0: (default .5) Fraction of individuals in population 0 (the rest will be in population 1)
    :type freq_pop_0: float

    :param chr_count: (default one chromosome per SNP) Number of chromosomes to which SNPs should be assigned. The SNPs will
     be assigned as evenly as possible. Chromosome names are integers starting with 1. SNP positions within a chromosome are sequential
     integers starting with 1.
    :type chr_count: int

    :rtype: :class:`.SnpData`

    :Example:

    >>> snpdata = snp_gen(fst=.1,dfr=.5,iid_count=200,sid_count=20,maf_low=.05,seed=6)
    >>> print(int(snpdata.iid_count), int(snpdata.sid_count)) #because of rounding got 190 individuals
    190 20

    :Also See: :class:`.snpreader.SnpGen`

    """
    from pysnptools.snpreader import SnpData

    assert 0 <= freq_pop_0 <= 1.0, "assert 0 <= freq_pop_0 <= 1.0"

    if seed is not None:
        np.random.seed(int(seed % 2147483647))  # old maxint

    iid_solo_count = iid_count - iid_count * dfr
    family_count = int(iid_count * dfr / (2 * sibs_per_family))

    ancestral = np.random.uniform(
        maf_low, maf_high, sid_count
    )  # sample ancestral allele frequencies

    snp_list = []
    for population_index, freq_pop in enumerate([freq_pop_0, 1.0 - freq_pop_0]):
        logging.info("Simulating SNPs from a population %i" % population_index)
        snps_parents = _generate_snps(
            ancestral, fst, int(iid_solo_count * freq_pop), sid_count
        )
        snp_list.append(snps_parents)
        snp_list.append(
            _generate_kids(
                parent_snps=snps_parents,
                family_count=int(freq_pop * family_count),
                sibs_per_family=sibs_per_family,
            )
        )

    snp_list.append(
        _generate_kids(
            parent_snps=np.concatenate(snp_list),
            family_count=family_count,
            sibs_per_family=sibs_per_family,
        )
    )
    val = np.concatenate(snp_list)

    if not label_with_pop:
        iid = np.array(
            [
                ["f_{0}".format(iid_index), "i_{0}".format(iid_index)]
                for iid_index in range(val.shape[0])
            ],
            dtype=str,
        ).reshape(-1, 2)
    else:
        assert len(snp_list) == 5, "real assert"
        iid0 = [
            ["0", str(iid_index)]
            for iid_index in range(len(snp_list[0]) + len(snp_list[1]))
        ]  # parents and children of pop 0
        iid1 = [
            ["1", str(iid_index)]
            for iid_index in range(len(snp_list[2]) + len(snp_list[3]))
        ]  # parents and children of pop 1
        iid2 = [
            ["2", str(iid_index)] for iid_index in range(len(snp_list[4]))
        ]  # children with parents in any pop
        iid = np.array(iid0 + iid1 + iid2, dtype=str).reshape(-1, 2)

    sid = np.array(
        ["snp_{0}".format(sid_index) for sid_index in range(val.shape[1])], dtype="str"
    )

    if chr_count is None:
        chr_count = len(sid)

    assert (
        len(sid) == 0 or chr_count > 0
    ), "chr_count must be at least 1 (unless sid_count is 0)"
    sid_per_chrom = int(np.ceil(float(len(sid)) / max(1, chr_count)))
    pos = np.array(
        list(
            [
                1 + sid_index // sid_per_chrom,
                1 + sid_index % sid_per_chrom,
                1 + sid_index % sid_per_chrom,
            ]
            for sid_index in range(len(sid))
        )
    )
    if len(sid) == 0:  # make it work when no sids are wanted
        pos = pos.reshape(len(sid), 3)

    snpdata = SnpData(
        iid=iid,
        sid=sid,
        val=val,
        pos=pos,
        parent_string="snp_gen(fst={0}, dfr={1}, iid_count={2}, sid_count={3}, maf_low={4}, maf_high={5}, seed={6}, sibs_per_family={7}, freq_pop_0={8})".format(
            fst,
            dfr,
            iid_count,
            sid_count,
            maf_low,
            maf_high,
            seed,
            sibs_per_family,
            freq_pop_0,
        ),
    )

    # if snpdata.iid_count != iid_count:
    #    logging.warn("Because of rounding the actual number of iids is {0} rather than the requested {1}".format(snpdata.iid_count, iid_count))

    return snpdata


def _beta_mode(a, f):  # !!!
    alpha, beta = (a * (1.0 - f) / f, (1.0 - a) * (1.0 - f) / f)
    print(alpha, beta)
    mean = alpha / (alpha + beta)
    var = alpha * beta / ((alpha + beta) ** 2 * (alpha + beta + 1))
    return mean, var**0.5


def _generate_snps(ancestral, fst, iid_count, sid_count):
    """
    Generates genotypes with a certain MAF and optionally with population structure.
    In case of no population structure, they are sampled from a binomial,
    otherwise from a Beta-Binomial (Balding and Nichols, 1995).
    """
    if fst == 0.0:
        alpha = ancestral  # special treatment if no population structure
    else:
        alpha = np.random.beta(
            ancestral * (1.0 - fst) / fst,
            (1.0 - ancestral) * (1.0 - fst) / fst,
            sid_count,
        )

    # generate from population frequencies
    snps = np.zeros(
        (iid_count, sid_count), dtype="int8"
    )  # .zeros not .empty because will be adding 1's to it
    for i in range(2):  # "2" for diploid
        # sample each allele
        rand = np.random.random((iid_count, sid_count))
        snps[rand < alpha] += 1
    return snps


def _generate_kids(
    parent_snps, family_count, sibs_per_family
):  # !!! should it be sibs, kids, or children
    """
    generate a single set of family members
    """
    parent_count, sid_count = parent_snps.shape
    assert parent_count >= 2 * family_count, "parent_count>=2*family_count"

    parent_permutation = np.random.permutation(parent_count)
    snps = np.zeros((family_count * sibs_per_family, sid_count), dtype=np.float64)
    for copy_index in range(2):  # "2" for diploid
        sample = parent_snps[
            parent_permutation[
                copy_index * family_count : (copy_index + 1) * family_count
            ],
            :,
        ]  # sample each allele
        for kid_index in range(sibs_per_family):
            rand = np.random.random((family_count, sid_count))
            snps[kid_index * family_count : (kid_index + 1) * family_count][
                rand < 0.5 * sample
            ] += 1
    return snps


def _encode_snp(entry):
    if entry == 0:
        return "A A"
    elif entry == 1:
        return "A C"
    elif entry == 2:
        return "C C"


def _generate_phenotype(snp_data, causals, genetic_var, noise_var, seed=None):
    """
    generate phenotype given genotype

    'causals' can be either an array of indexes to the causal snps or the number of causal snps desired.
    """

    if seed is not None:
        np.random.seed(
            int(seed % 2147483647)
        )  # !!!Better to use numpy.random.RandomState instead (look for other places in code, too) #old maxint

    try:
        num_causal = len(causals)
        causal_idx = causals
    except Exception:
        num_causal = causals
        # the "if..else" is a work around because the linux version of np.random.choice doesn't like to select zero items from an empty list. We need to call random in either case so that the random seed ends up in the expected state
        causal_idx = np.random.choice(
            np.arange(snp_data.sid_count if num_causal > 0 else 1),
            size=num_causal,
            replace=False,
        )

    num_phenotypes = 1
    mean = 0.0
    X = snp_data.val.copy()
    X.flags.writeable = False
    X_causal = X[:, causal_idx]
    X_causal = 1.0 / np.sqrt(X_causal.shape[1]) * X_causal
    W = (
        np.random.randn(num_causal, num_phenotypes) * np.sqrt(genetic_var) + mean
    )  # Weight of each causal SNP
    XW = np.dot(X_causal, W)
    noise_std = np.sqrt(noise_var)
    Z = noise_std * np.random.randn(X_causal.shape[0], num_phenotypes)
    y = XW + Z
    y = y[:, 0]

    return y


class TestGenerate(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.currentFolder = os.path.dirname(os.path.realpath(__file__))

    def gen_and_compare(self, output_file, **kwargs):
        from pysnptools.snpreader import Bed

        gen_snpdata = snp_gen(**kwargs)
        # pstutil.create_directory_if_necessary(self.currentFolder + "/tempdir/" + output_file,isfile=True)
        # Bed.write(gen_snpdata, self.currentFolder + "/tempdir/" + output_file)  #comment out
        bed = Bed(
            self.currentFolder + "/../../tests/datasets/generate/" + output_file,
            count_A1=False,
        )
        ref_snpdata = bed.read()
        assert gen_snpdata == ref_snpdata, "Failure on " + output_file
        return gen_snpdata
        # !!! Ped doesn't seem to round trip well
        # !!! Hdf5 doesn't seem to round trip well

    def test_gen1(self):
        self.gen_and_compare(
            "gen1.bed",
            fst=0,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
        )

    def test_gen2(self):
        self.gen_and_compare(
            "gen2.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
        )

    def test_gen2b(self):
        """
        Test that different seed produces different result
        """
        from pysnptools.snpreader import Bed

        gen_snpdata = self.gen_and_compare(
            "gen2b.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=6,
        )
        ref_snpdata = Bed(
            self.currentFolder + "/../../tests/datasets/generate/gen2.bed",
            count_A1=False,
        ).read()
        assert (
            gen_snpdata != ref_snpdata
        ), "Expect different seeds to produce different results"

    def test_gen3(self):
        self.gen_and_compare(
            "gen3.bed",
            fst=0.1,
            dfr=0,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
        )

    def test_gen4(self):
        self.gen_and_compare(
            "gen4.bed",
            fst=0.1,
            dfr=0.01,
            iid_count=200,
            sid_count=20,
            maf_low=0.1,
            seed=5,
        )

    def test_gen5(self):
        from pysnptools.snpreader import Bed

        gen_snpdata = self.gen_and_compare(
            "gen5.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            maf_high=0.4,
            seed=5,
        )
        ref_snpdata = Bed(
            self.currentFolder + "/../../tests/datasets/generate/gen2.bed",
            count_A1=False,
        ).read()
        assert (
            gen_snpdata != ref_snpdata
        ), "Expect different seeds to produce different results"

    def test_gen6(self):
        from pysnptools.snpreader import Bed

        gen_snpdata = self.gen_and_compare(
            "gen6.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
            sibs_per_family=5,
        )
        ref_snpdata = Bed(
            self.currentFolder + "/../../tests/datasets/generate/gen2.bed",
            count_A1=False,
        ).read()
        assert (
            gen_snpdata != ref_snpdata
        ), "Expect different seeds to produce different results"

    def test_gen7(self):
        from pysnptools.snpreader import Bed

        gen_snpdata = self.gen_and_compare(
            "gen7.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
            freq_pop_0=0.75,
        )
        ref_snpdata = Bed(
            self.currentFolder + "/../../tests/datasets/generate/gen2.bed",
            count_A1=False,
        ).read()
        assert gen_snpdata != ref_snpdata

    def test_gen8(self):
        self.gen_and_compare(
            "gen8a.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
            chr_count=3,
        )
        self.gen_and_compare(
            "gen8b.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
            chr_count=4,
        )
        self.gen_and_compare(
            "gen8c.bed",
            fst=0.1,
            dfr=0.5,
            iid_count=200,
            sid_count=20,
            maf_low=0.05,
            seed=5,
            chr_count=6,
        )

    def test_pheno1(self):
        from pysnptools.snpreader import Bed, SnpData, SnpNpz

        some_snp_data = Bed(
            self.currentFolder + "/../../tests/datasets/generate/gen2.bed",
            count_A1=False,
        ).read()
        gen_snpdata = SnpData(
            iid=some_snp_data.iid,
            sid=["pheno"],
            val=_generate_phenotype(
                some_snp_data, 10, genetic_var=0.5, noise_var=0.5, seed=5
            ).reshape(-1, 1),
        )
        # SnpNpz.write(r'c:\deldir\pheno1.snp.npz',gen_snpdata)
        ref_snpdata = SnpNpz(
            self.currentFolder + "/../../tests/datasets/generate/pheno1.snp.npz"
        ).read()
        assert gen_snpdata == ref_snpdata

    def test_gensmall(self):
        # Just checking that doesn't generate errors
        for iid_count in [10, 5, 3, 2, 1, 0]:
            for sid_count in [0, 10, 5, 3, 2, 1]:
                for chr_count in [30, 10, 5, 3, 2, 1, 0]:
                    if chr_count == 0 and sid_count > 0:
                        continue  # not break
                    logging.debug(
                        "{0}, {1}, {2}".format(iid_count, sid_count, chr_count)
                    )
                    snpdata = snp_gen(
                        fst=0.1,
                        dfr=0.5,
                        iid_count=iid_count,
                        sid_count=sid_count,
                        maf_low=0.05,
                        seed=6,
                        chr_count=chr_count,
                    )
                    assert snpdata.iid_count <= iid_count
                    assert snpdata.sid_count == sid_count
                    assert len(snpdata.pos) == 0 or max(snpdata.pos[:, 0]) <= chr_count
                    assert len(snpdata.pos) == 0 or max(snpdata.pos[:, 1]) <= int(
                        max(1, np.ceil(float(sid_count) / chr_count))
                    )
                    assert len(snpdata.pos) == 0 or max(snpdata.pos[:, 2]) <= int(
                        max(1, np.ceil(float(sid_count) / chr_count))
                    )

    def test_doc_test(self):
        import pysnptools.util.generate as gen_mod

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)) + "/..")
        result = doctest.testmod(gen_mod, optionflags=doctest.ELLIPSIS)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestGenerate))

    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    print("done")



import re
import os
import psutil
import logging
from contextlib import contextmanager
import subprocess
import random
import datetime
import shutil

from pysnptools.util.filecache import FileCache, _DibLib
from pysnptools.util import format_delta, _mbps


class PeerToPeer(FileCache):
    """
    A :class:`.FileCache` that allows multiple machines (or processes on one machine) to share files in a peer-to-peer fashion.

    See :class:`.FileCache` for general examples of using FileCache.

    This is useful for a cluster with multiple machines with persistent storage and a fast network. Every time a machine writes a file,
    the file stays on that machine, but information about the file's name and location is saved to a common directory. When a new
    machine needs the file, it looks to the common directory for a location and then copies from that location. That second location
    is also saved to the common directory. Later machines that need the file will randomly pick among the available locations. When a file is removed, it
    is only removed from the common directory; clean up of local copies happens later when space is needed for other reads.


    **Constructor:**
        :Parameters: * **common_directory** (:class:`.FileCache` or *string*) -- The path or :class:`.FileCache` where
                       each file's name and locations (but not contents) are stored. This path must be read/writable by every machine in the cluster.
                     * **id_and_path_function** (*function*) -- A function (or lambda) that deterministically returns two values: 1. an id unique to each machine
                       (or process) and 2. a file path to this machine's local storage that any machine
                       in the cluster can read.
                     * **leave_space** (*number*) -- (Default "0") How much free space must be left on local drive after downloads and writes.

        :Example:

        >>> #Suppose that every machine can access a shared 'peertopeer1' and that 'peertopeer1/IPADDRESS' is local to each machine.
        >>> from pysnptools.util.filecache import PeerToPeer, ip_address
        >>> def id_and_path_function():
        ...     ip = ip_address()
        ...     return ip, 'peertopeer1/{0}'.format(ip)
        >>> file_cache = PeerToPeer(common_directory='peertopeer1/common',id_and_path_function=id_and_path_function)
        >>> file_cache.rmtree()
        >>> file_cache.save('sub1/file1.txt','Hello')
        >>> file_cache.file_exists('sub1/file1.txt')
        True

    """

    # Later features:
    #   Limit the amount of copying for a given file
    #   Work even when multiple tasks on same node ask for the same file
    #   Get open_read,close to work with python With statement
    #   Have a way to clean up lost files
    def __init__(self, common_directory, id_and_path_function, leave_space=0):
        from pysnptools.util.filecache import LocalCache

        super(PeerToPeer, self).__init__()
        shared_directory_str = str(common_directory)
        if isinstance(common_directory, str):
            common_directory = LocalCache(common_directory)
        self.common_directory = common_directory
        self.id_and_path_function = id_and_path_function
        self.leave_space = leave_space
        assert not self.common_directory._simple_file_exists(
            "main.txt"
        ), "A common_directory cannot exist where a file already exists."
        self._str = "{0}('{1}',id_and_path_function=...{2}')".format(
            self.__class__.__name__,
            shared_directory_str,
            "" if self.leave_space == 0 else ",leave_space={0}".format(leave_space),
        )
        # !!!test  if self.leave_space==0
        # !!!test on one machine and multiple processes

    def __repr__(self):
        return self._str

    @property
    def name(self):
        """
        A path-like name for this `PeerToPeer`.

        :rtype: string
        """
        return self.common_directory.name

    def _simple_file_exists(self, simple_file_name):
        return self.common_directory.file_exists(simple_file_name + "/main.txt")

    def _far_file_sequence(self, dir_path):
        storage_list = [f for f in dir_path.walk() if self._copy_main_pattern.match(f)]
        random.shuffle(storage_list)

        for storage_item in storage_list:
            storage_path = dir_path.load(storage_item)
            yield storage_path, len(storage_list)

    _copy_main_pattern = re.compile(r"^((main)|(copy_.*))\.txt$")
    _copy_pattern = re.compile(r"^copy_.*\.txt$")

    def _copy_time_stamp(self, storage_path, local_path):
        timestamp = os.path.getmtime(storage_path)
        with open(
            local_path, "a"
        ):  # Touch the local file so that its mod time will be later than the remote time (see http://stackoverflow.com/questions/1158076/implement-touch-using-python)
            os.utime(local_path, (timestamp, timestamp))

    @contextmanager
    def _simple_open_read(self, simple_file_name, updater=None):
        import pysnptools.util as pstutil  # put here to avoid recursive nesting

        logging.debug("open_read('{0}')".format(simple_file_name))
        #
        # Returns name of local file and locks that local until released
        #

        # Try to Get locally, return it if worked
        # Is there enough room? If not delete some files. If still not, fail
        # Ask origin for randomized list of copies (the last one will be the origin, but not all have to be included)
        # Try getting each copy. If all fail, then fail
        # Register local copy with the origin
        dir_path = self.common_directory.join(simple_file_name)
        unique_name, root = self.id_and_path_function()
        local_path = root + "/" + simple_file_name
        pstutil.create_directory_if_necessary(local_path, isfile=True)
        copy_name = "copy_{0}.txt".format(unique_name)
        main_path = self._robust_load_main(dir_path)
        file_size = os.path.getsize(main_path)

        if os.path.exists(local_path):
            if main_path == local_path or dir_path.file_exists(copy_name):
                logging.info("\tfound local")
                assert file_size == os.path.getsize(
                    local_path
                ), "Local file doesn't have the same size as the main file"
                yield local_path
                return
            else:
                logging.info("\tlost local found. Will remove")
                os.remove(local_path)
                pstutil.create_directory_if_necessary(local_path, isfile=True)

        dib_lib = _DibLib(unique_name, dir_path, dir_path, "dibs")
        try:
            dib_lib.wait_for_turn()

            # Now we can copy. Choose a source at random
            for storage_path, storage_count in self._far_file_sequence(
                dir_path
            ):  # In a loop because in the future may want to handle copies being deleted.
                try:  # If something goes wrong, try the next one
                    self._net_use(storage_path)
                    assert file_size == os.path.getsize(
                        storage_path
                    ), "File to copy from doesn't have the same size as the main file"
                    if (
                        psutil.disk_usage(os.path.dirname(local_path)).free - file_size
                        < self.leave_space
                    ):  # clean up the directories
                        # Check every local file and if it is not in the directory (e.g. a temp file from other programs), remove it
                        local_dir = os.path.split(local_path)[0]
                        for other_file in os.listdir(local_dir):
                            full_file = local_dir + "/" + other_file
                            if os.path.isfile(
                                full_file
                            ) and not self._simple_file_exists(other_file):
                                logging.info(
                                    "\tNeed space and file isn't in directory so removing it. '{0}'".format(
                                        full_file
                                    )
                                )
                                os.remove(full_file)
                    logging.info(
                        "\tshutil.copyfile('{0}','{1}')".format(
                            storage_path, local_path
                        )
                    )
                    then = datetime.datetime.now()
                    shutil.copyfile(storage_path, local_path)
                    shutil.copystat(storage_path, local_path)
                    self._copy_time_stamp(storage_path, local_path)
                    dir_path.save(copy_name, local_path)
                    delta_sec = max(
                        (datetime.datetime.now() - then).total_seconds(), 1.0
                    )
                    try:  # The 'try' stops this logging message from getting a div by zero error some times
                        logging.info(
                            "Copy time is {0}. Copy speed is {1} Mbps".format(
                                delta_sec, _mbps(file_size, delta_sec)
                            )
                        )
                    except Exception:
                        logging.info(
                            "Copy time is {0}. Copy speed can't be calculated Mbps".format(
                                format_delta(delta_sec)
                            )
                        )
                    break
                except Exception as e:
                    if os.path.exists(local_path):
                        logging.warning(
                            "If a local file was created, but something went wrong (e.g. the source disappeared part way through the copying), so we remove it"
                        )
                        os.remove(local_path)
                    logging.warning("Ignore exception {0}".format(e))
            assert file_size == os.path.getsize(
                local_path
            ), "File did not copy or did not copy completely."

        finally:
            dib_lib.remove_dibs()

        yield local_path

        assert dir_path._simple_file_exists(
            "main.txt"
        ), "File doesn't exist: '{0}'".format(dir_path)

    @staticmethod
    def _net_use(path):
        if not path.startswith(r"\\"):
            return
        path = path.replace("/", "\\")
        path = "\\".join(path.split("\\")[:4])
        if True:
            subprocess.call(r'net use {0} "" /u:""'.format(path), shell=True)
        else:
            process = subprocess.Popen(
                r'net use {0} "" /u:""'.format(path),
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, err = process.communicate()  # wait for the process to terminate
            if (
                "Multiple connections to a server or shared resource by the same user, using more than one user name, are not allowed."
                in err
                or "System error 5" in err
            ):
                return
            logging.warning(err)
            # errcode = process.returncode

    def _robust_load_main(self, dir_path):
        main_path = dir_path.load("main.txt")
        try:
            self._net_use(main_path)
        except Exception as e:
            logging.info(
                "ignoring exception when doing net use on {0}, {1}".format(
                    main_path, e.message
                )
            )
        if os.path.exists(main_path):
            return main_path
        logging.info(
            "'{0}' does not exist, so looking for other copies.".format(main_path)
        )
        for copy_file in dir_path._simple_walk():
            if not self._copy_pattern.match(copy_file):
                continue
            copy_path = dir_path.load(copy_file)
            try:
                self._net_use(copy_path)
            except Exception as e:
                logging.info(
                    "ignoring exception when doing net use on {0}, {1}".format(
                        copy_path, e.message
                    )
                )
            if os.path.exists(copy_path):
                logging.info("Doing a copy to main repair. Not sure if this is safe")
                dir_path._simple_remove("main.txt")
                dir_path.save("main.txt", copy_path)
                dir_path._simple_remove(copy_file)
                return copy_path
        raise Exception(
            "The main copy of the file can't be found and no replacement is available. '{0}'".format(
                main_path
            )
        )

    def _remove_local_if_any(self, path):
        if path is None:
            unique_name, root = self.id_and_path_function()
            if os.path.exists(root):
                shutil.rmtree(root)
        else:
            directory, simple_file_name = self._split(path)
            unique_name, root = directory.id_and_path_function()
            local_path = root + "/" + simple_file_name
            if os.path.exists(local_path):
                logging.info("Removing '{0}' to save space".format(local_path))
                os.remove(local_path)

    @contextmanager
    def _simple_open_write(self, simple_file_name, size=0, updater=None):
        logging.info("open_write('{0}',size={1})".format(simple_file_name, size))
        # Register the file name in the directory
        dir_path = self.common_directory.join(simple_file_name)
        assert not dir_path._simple_file_exists(
            "main.txt"
        ), "Can't open a file for write if it already exists."
        assert not self._at_least_one(
            dir_path._simple_walk()
        ), "Can't open a file for write if a directory with files already has the same name ({0},{1})".format(
            self, simple_file_name
        )
        unique_name, root = self.id_and_path_function()
        local_path = root + "/" + simple_file_name
        # Anything in storage (file or directory) can be cleaned up.
        self._create_directory(local_path)
        if os.path.exists(local_path):
            logging.info(
                "File with the same name found locally, but without a directory entry, so removing it. '{0}'".format(
                    local_path
                )
            )
            os.remove(local_path)
        assert (
            psutil.disk_usage(os.path.dirname(local_path)).free - size
            > self.leave_space
        ), "Not enough space for '{0}. Need to make space".format(local_path)

        yield local_path

        logging.info("close('{0}')".format(simple_file_name))
        assert os.path.exists(local_path), "Expect file at '{0}'".format(local_path)
        assert (
            psutil.disk_usage(os.path.dirname(local_path)).free > self.leave_space
        ), "Not enough space for '{0}'. Need to make space".format(local_path)

        dir_path.save("main.txt", local_path)

    def _remove_internal(self, path):
        for storage_item in self.common_directory.walk(path):
            storage_path = self.common_directory.load(storage_item)
            if (
                storage_path != ""
            ):  # Dibs files don't point anywhere, so we don't need to delete the file they point to.
                try:
                    logging.debug("/tos.remove('{0}')".format(storage_path))
                    os.remove(
                        storage_path
                    )  # In storage, we remove the files, but not the folder
                except Exception:
                    logging.debug(
                        "Can't remove (because it's not there): '{0}'".format(
                            storage_path
                        )
                    )
        self.common_directory.rmtree(path)

    def _simple_rmtree(self, updater=None):
        logging.info("rmtree -- {0}".format(self.common_directory))
        self._remove_internal(None)

    def _simple_remove(self, simple_file_name, updater=None):
        assert self._simple_file_exists(
            simple_file_name
        ), "Expect file to exist (and be a file)"
        self._remove_internal(simple_file_name)

    def _simple_getmtime(self, simple_file_name):
        return self.common_directory._simple_getmtime(simple_file_name + "/main.txt")

    def _simple_join(self, path):
        def closure():
            unique_name, root = self.id_and_path_function()
            return unique_name, root + "/" + path

        fs = PeerToPeer(
            self.common_directory._simple_join(path),
            id_and_path_function=closure,
            leave_space=self.leave_space,
        )
        return fs

    def _simple_walk(self):
        for storage_item in self.common_directory._simple_walk():
            head, tail = os.path.split(storage_item)
            if tail == "main.txt":
                yield head


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.pstreader import PstReader
import pysnptools.util as pstutil


def _default_empty_creator(count):
    return np.empty([count or 0, 0], dtype="str")


def _default_empty_creator_val(row_count, col_count):
    return np.empty([row_count, col_count], dtype="str")


class PstData(PstReader):
    """A :class:`.PstReader` for holding values in-memory, along with related row and col information.
    It is usually created by calling the :meth:`.PstReader.read` method on another :class:`.PstReader`, for example, :class:`.PstNpz`. It can also be constructed directly.

    See :class:`.PstReader` for details and examples.

    **Constructor:**
        :Parameters: * **row** (an array of anything) -- The :attr:`PstReader.row` information
                     * **col** (an array of anything) -- The :attr:`PstReader.col` information
                     * **val** (a 2-D array of floats or an array of floats) -- The values
                     * **row_property** (optional, an array of anything) -- Additional information associated with each row.
                     * **col_property** (optional, an array of anything) -- Additional information associated with each col.
                     * **name** (optional, string) -- Information to be display about the origin of this data
                     * **copyinputs_function** (optional, function) -- *Used internally by optional clustering code*

        :Example:

        >>> from pysnptools.pstreader import PstData
        >>> pstdata = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> print(pstdata.val[0,1], pstdata.row_count, pstdata.col_count)
        2.0 2 3

    **Equality:**

        Two PstData objects are equal if their five arrays (:attr:`PstData.val`, :attr:`PstReader.row`, :attr:`PstReader.col`, :attr:`PstReader.row_property`, and :attr:`PstReader.col_property`) arrays are equal.
        (Their 'name' does not need to be the same).
        If either :attr:`PstData.val` contains NaN, the objects will not be equal. However, :meth:`.PstData.allclose` can be used to treat NaN as regular values.
        Any NaN's in the other four arrays are treated as regular values.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.pstreader import PstData
        >>> pstdata1 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> pstdata2 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> print(pstdata1 == pstdata2) #True, because all the arrays have the same values.
        True
        >>> print(pstdata1.val is pstdata2.val) #False, because the two arrays have different memory.
        False
        >>> pstdata3 = PstData(row=['a','b'], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> pstdata4 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> print(pstdata3 == pstdata4) #False, because the rows have different ids.
        False
        >>> pstdata5 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]])
        >>> pstdata6 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]])
        >>> print(pstdata5 == pstdata6) #False, because the val's contain NaN
        False
        >>> print(pstdata5.allclose(pstdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True

    **Methods beyond** :class:`.PstReader`


    """

    def __init__(
        self,
        row,
        col,
        val,
        row_property=None,
        col_property=None,
        name=None,
        parent_string=None,
        copyinputs_function=None,
    ):
        super(PstData, self).__init__()

        self._val = None

        self._row = PstData._fixup_input(row)
        self._col = PstData._fixup_input(col)
        if self._row.dtype == self._col.dtype and np.array_equal(
            self._row, self._col
        ):  # If it's square, mark it so by making the col and row the same object
            self._col = self._row
        self._row_property = PstData._fixup_input(row_property, count=len(self._row))
        self._col_property = PstData._fixup_input(col_property, count=len(self._col))
        self._val = PstData._fixup_input_val(
            val, row_count=len(self._row), col_count=len(self._col)
        )
        self._assert_row_col_and_properties(check_val=True)

        name = name or parent_string or ""
        if parent_string is not None:
            warnings.warn(
                "'parent_string' is deprecated. Use 'name'", DeprecationWarning
            )
        self._name = name

    def _assert_row_col_and_properties(self, check_val):
        if check_val:
            assert self._val.shape[:2] == (
                len(self._row),
                len(self._col),
            ), "val shape should match that of row_count x col_count"

    def __eq__(a, b):
        return a.allclose(b, equal_nan=False)

    def allclose(self, value, equal_nan=True):
        """
        :param value: Other object with which to compare.
        :type value: :class:`PstData`
        :param equal_nan: (Default: True) Tells if NaN in :attr:`PstData.val` should be treated as regular values when testing equality.
        :type equal_nan: bool

        >>> import numpy as np
        >>> pstdata5 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]])
        >>> pstdata6 = PstData(row=[['fam0','iid0'],['fam0','iid1']], col=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]])
        >>> print(pstdata5.allclose(pstdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True
        >>> print(pstdata5.allclose(pstdata6,equal_nan=False)) #False, if we consider the NaN as special values, all the arrays are not equal.
        False

        """
        try:
            return (
                PstData._allclose(self.row, value.row, equal_nan=True)
                and PstData._allclose(self.col, value.col, equal_nan=True)
                and PstData._allclose(
                    self.row_property, value.row_property, equal_nan=True
                )
                and PstData._allclose(
                    self.col_property, value.col_property, equal_nan=True
                )
                and np.allclose(self.val, value.val, equal_nan=equal_nan)
            )
        except Exception:
            return False

    @staticmethod
    def _allclose(a, b, equal_nan=True):
        if not equal_nan:
            return np.array_equal(a, b)
        if a.dtype == "float" and b.dtype == "float":
            return np.allclose(a, b, equal_nan=equal_nan)
        return np.array_equal(a, b)

    @staticmethod
    def _fixup_input(
        input, count=None, empty_creator=_default_empty_creator, dtype=None
    ):
        if input is None or len(input) == 0:
            input = empty_creator(count)
        elif not isinstance(input, np.ndarray):
            input = np.array(input, dtype=dtype)

        assert (
            count is None or len(input) == count
        ), "Expect length of {0} for input {1}".format(count, input)

        return input

    @staticmethod
    def _fixup_input_val(
        input,
        row_count,
        col_count,
        empty_creator=_default_empty_creator_val,
        _require_float32_64=True,
        xp=None,
    ):
        xp = pstutil.array_module(xp)

        if input is None:
            assert (
                row_count == 0 or col_count == 0
            ), "If val is None, either row_count or col_count must be 0"
            input = _default_empty_creator_val(row_count, col_count)
        elif not isinstance(input, xp.ndarray):
            input = xp.array(input, dtype=xp.float64)
        elif _require_float32_64 and input.dtype not in [xp.float32, xp.float64]:
            input = xp.array(input, dtype=xp.float64)

        assert len(input.shape) in {2, 3}, "Expect val to be two or three dimensional."
        assert (
            input.shape[0] == row_count
        ), "Expect number of rows ({0}) in val to match the number of row names given ({1})".format(
            input.shape[0], row_count
        )
        assert (
            input.shape[1] == col_count
        ), "Expect number of columns ({0}) in val to match the number of column names given ({1})".format(
            input.shape[1], col_count
        )

        return input

    def __repr__(self):
        if self._name == "":
            return "{0}()".format(self.__class__.__name__)
        else:
            return "{0}({1})".format(self.__class__.__name__, self._name)

    def copyinputs(self, copier):
        pass

    @property
    def row(self):
        return self._row

    @property
    def col(self):
        return self._col

    @property
    def row_property(self):
        return self._row_property

    @property
    def col_property(self):
        return self._col_property

    @property
    def val_shape(self):
        """
        Tells the shape of value for a given individual and SNP. If None, means a single scalar value.
        """
        if len(self._val.shape) == 2:
            return None
        else:
            return self._val.shape[2]

    @property
    def val(self):
        """The 2D NumPy array of floats (or array of floats) that represents the values.  You can get or set this property.

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file
        >>>
        >>> pstnpz_file = example_file('tests/datasets/little.pst.npz')
        >>> pstdata = PstNpz(pstnpz_file)[:5,:].read() #read data for first 5 rows
        >>> print(pstdata.val[4,5]) #print one of the values
        2.0
        """
        return self._val

    @val.setter
    def val(self, new_value):
        self._val = PstData._fixup_input_val(
            new_value, row_count=len(self._row), col_count=len(self._col)
        )
        self._assert_row_col_and_properties(check_val=True)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        dtype = np.dtype(dtype)
        val, shares_memory = self._apply_sparray_or_slice_to_val(
            self.val,
            row_index_or_none,
            col_index_or_none,
            order,
            dtype,
            force_python_only,
            num_threads,
        )
        if shares_memory and not view_ok:
            val = val.copy(order="K")
        return val


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import numpy as np
from itertools import *
import pandas as pd
import logging
import time
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData


class _Dist2Snp(SnpReader):
    def __init__(self, snpreader, max_weight=2.0, block_size=None):
        super(_Dist2Snp, self).__init__()

        self.distreader = snpreader
        self.max_weight = max_weight
        self.block_size = block_size

    @property
    def row(self):
        return self.distreader.iid

    @property
    def col(self):
        return self.distreader.col

    def __repr__(self):
        return self._internal_repr()

    def _internal_repr(self):  # !!! merge this with __repr__
        s = "{0}.as_snp(".format(self.distreader)
        if self.block_size is not None:
            s += "block_size={0}".format(self.block_size)
        s += ")"
        return s

    def copyinputs(self, copier):
        # Doesn't need run_once
        copier.input(self.distreader)

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        from pysnptools.distreader import DistReader

        dtype = np.dtype(dtype)

        assert (
            row_index_or_none is None and col_index_or_none is None
        )  # real assert because indexing should already be pushed to the inner distreader

        weights = np.array([0, 0.5, 1], dtype=dtype) * self.max_weight

        # Do all-at-once (not in blocks) if 1. No block size is given or 2. The #ofSNPs < Min(block_size,iid_count)
        if self.block_size is None or (
            self.sid_count <= self.block_size or self.sid_count <= self.iid_count
        ):
            distdata = DistReader._as_distdata(
                self.distreader,
                dtype=dtype,
                order=order,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )
            val = (distdata.val * weights).sum(axis=-1)
            has_right_order = order = (
                "A"
                or (order == "C" and val.flags["C_CONTIGUOUS"])
                or (order == "F" and val.flags["F_CONTIGUOUS"])
            )
            assert has_right_order
            return val
        else:  # Do in blocks
            t0 = time.time()
            if order == "A":
                order = "F"
            val = np.zeros(
                [self.iid_count, self.sid_count], dtype=dtype, order=order
            )  # LATER should use empty or fillnan

            logging.info(
                "reading {0} distribution data in blocks of {1} SNPs and finding expected values (for {2} individuals)".format(
                    self.sid_count, self.block_size, self.iid_count
                )
            )
            ct = 0
            ts = time.time()

            for start in range(0, self.sid_count, self.block_size):
                ct += self.block_size
                distdata = self.distreader[:, start : start + self.block_size].read(
                    order=order,
                    dtype=dtype,
                    force_python_only=force_python_only,
                    view_ok=True,
                    num_threads=num_threads,
                )  # a view is always OK, because we'll allocate memory in the next step
                val[:, start : start + self.block_size] = (distdata.val * weights).sum(
                    axis=-1
                )
                if ct % self.block_size == 0:
                    diff = time.time() - ts
                    if diff > 5:
                        logging.info("read %s SNPs in %.2f seconds" % (ct, diff))

            t1 = time.time()
            logging.info("%.2f seconds elapsed" % (t1 - t0))

            return val

    def __getitem__(self, iid_indexer_and_snp_indexer):
        row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        return _Dist2Snp(
            self.distreader[row_index_or_none, col_index_or_none],
            max_weight=self.max_weight,
            block_size=self.block_size,
        )

    @property
    def sid(self):
        """The :attr:`SnpReader.sid` property of the SNP data."""
        return self.distreader.sid

    @property
    def sid_count(self):
        """The :attr:`SnpReader.sid_count` property of the SNP data."""
        return self.distreader.sid_count

    @property
    def pos(self):
        """The :attr:`SnpReader.pos` property of the SNP data."""
        return self.distreader.pos


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod()
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import logging
import os
import shutil
import numpy as np
import unittest
import doctest
import pysnptools.util as pstutil
from pysnptools.pstreader import PstData
from pysnptools.pstreader import PstMemMap
from pysnptools.distreader import DistReader, DistData
from pysnptools.util import log_in_place


class DistMemMap(PstMemMap, DistData):
    """
    A :class:`.DistData` that keeps its data in a memory-mapped file. This allows data large than fits in main memory.

    See :class:`.DistData` for general examples of using DistData.

    **Constructor:**
        :Parameters: **filename** (*string*) -- The *\*.dist.memmap* file to read.

        Also see :meth:`.DistMemMap.empty` and :meth:`.DistMemMap.write`.

        :Example:

        >>> from pysnptools.distreader import DistMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> mem_map_file = example_file("pysnptools/examples/tiny.dist.memmap")
        >>> dist_mem_map = DistMemMap(mem_map_file)
        >>> print(dist_mem_map.val[0,1], dist_mem_map.iid_count, dist_mem_map.sid_count)
        [0.43403135 0.28289911 0.28306954] 25 10

    **Methods inherited from** :class:`.DistData`

        :meth:`.DistData.allclose`

    **Methods beyond** :class:`.DistReader`

    """

    def __init__(self, *args, **kwargs):
        super(DistMemMap, self).__init__(*args, **kwargs)

    @property
    def val(self):
        """The 3D NumPy memmap array of floats that represents the distribution of SNP values. You can get this property, but cannot set it (except with itself)


        >>> from pysnptools.distreader import DistMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> mem_map_file = example_file("pysnptools/examples/tiny.dist.memmap")
        >>> dist_mem_map = DistMemMap(mem_map_file)
        >>> print(dist_mem_map.val[0,1])
        [0.43403135 0.28289911 0.28306954]
        """
        self._run_once()
        return self._val

    @val.setter
    def val(self, new_value):
        self._run_once()
        if self._val is new_value:
            return
        raise Exception("DistMemMap val's cannot be set to a different array")

    @property
    def offset(self):
        """The byte position in the file where the memory-mapped values start.

        (The disk space before this is used to store :attr:`DistReader.iid`, etc. information.
        This property is useful when interfacing with, for example, external Fortran and C matrix libraries.)

        """
        self._run_once()
        return self._offset

    @property
    def filename(self):
        """The name of the memory-mapped file"""
        # Don't need '_run_once'
        return self._filename

    @staticmethod
    def empty(iid, sid, filename, pos=None, order="F", dtype=np.float64):
        """Create an empty :class:`.DistMemMap` on disk.

        :param iid: The :attr:`DistReader.iid` information
        :type iid: an array of string pairs

        :param sid: The :attr:`DistReader.sid` information
        :type sid: an array of strings

        :param filename: name of memory-mapped file to create
        :type filename: string

        :param pos: optional -- The additional :attr:`DistReader.pos` information associated with each sid. Default: None
        :type pos: an array of numeric triples

        :param order: {'F' (default), 'C'}, optional -- Specify the order of the ndarray.
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`DistMemMap.val` ndarray.
        :type dtype: data-type

        :rtype: :class:`.DistMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.distreader import DistMemMap
        >>> filename = "tempdir/tiny.dist.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> dist_mem_map = DistMemMap.empty(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],filename=filename,order="F",dtype=np.float64)
        >>> dist_mem_map.val[:,:,:] = [[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                            [[0,1.,0],[0,.75,.25],[.5,.5,0]]]
        >>> dist_mem_map.flush()

        """

        self = DistMemMap(filename)
        self._empty_inner(
            row=iid,
            col=sid,
            filename=filename,
            row_property=None,
            col_property=pos,
            order=order,
            dtype=dtype,
            val_shape=3,
        )
        return self

    def flush(self):
        """Flush :attr:`DistMemMap.val` to disk and close the file. (If values or properties are accessed again, the file will be reopened.)

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.distreader import DistMemMap
        >>> filename = "tempdir/tiny.dist.memmap"
        >>> pstutil.create_directory_if_necessary(filename)
        >>> dist_mem_map = DistMemMap.empty(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'],filename=filename,order="F",dtype=np.float64)
        >>> dist_mem_map.val[:,:,:] = [[[.5,.5,0],[0,0,1],[.5,.5,0]],
        ...                            [[0,1.,0],[0,.75,.25],[.5,.5,0]]]
        >>> dist_mem_map.flush()

        """
        if self._ran_once:
            self.val.flush()
            del self._val
            self._ran_once = False

    @staticmethod
    def write(
        filename, distreader, order="A", dtype=None, block_size=None, num_threads=None
    ):
        """Writes a :class:`DistReader` to :class:`DistMemMap` format.

        :param filename: the name of the file to create
        :type filename: string
        :param distreader: The data that should be written to disk. It can also be any distreader, for example, :class:`.DistNpz`, :class:`.DistData`, or
           another :class:`.Bgen`.
        :type distreader: :class:`DistReader`
        :param order: {'A' (default), 'F', 'C'}, optional -- Specify the order of the ndarray. By default, will match the order of the input if knowable; otherwise, 'F'
        :type order: string or None
        :param dtype: {None (default), numpy.float64, numpy.float32}, optional -- The data-type for the :attr:`DistMemMap.val` ndarray.
             By default, will match the order of the input if knowable; otherwise np.float64.
        :type dtype: data-type
        :param block_size: The number of SNPs to read in a batch from *distreader*. Defaults to a *block_size* such that *block_size* \* *iid_count* is about 100,000.
        :type block_size: number
        :param num_threads: optional -- The number of threads with which to write data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int
        :rtype: :class:`.DistMemMap`

        >>> import pysnptools.util as pstutil
        >>> from pysnptools.distreader import Bgen, DistMemMap
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> distreader = Bgen(bgen_file)[:,:10] #Create a reader for the first 10 SNPs
        >>> pstutil.create_directory_if_necessary("tempdir/tiny.dist.memmap")
        >>> DistMemMap.write("tempdir/tiny.dist.memmap",distreader)      # Write distreader in DistMemMap format
        DistMemMap('tempdir/tiny.dist.memmap')

        """
        block_size = block_size or max((100_000) // max(1, distreader.row_count), 1)

        if hasattr(distreader, "val"):
            order = PstMemMap._order(distreader) if order == "A" else order
            dtype = dtype or distreader.val.dtype
        else:
            order = "F" if order == "A" else order
            dtype = dtype or np.float64
        dtype = np.dtype(dtype)

        self = PstMemMap.empty(
            distreader.row,
            distreader.col,
            filename + ".temp",
            row_property=distreader.row_property,
            col_property=distreader.col_property,
            order=order,
            dtype=dtype,
            val_shape=3,
        )
        if hasattr(distreader, "val"):
            self.val[:, :, :] = distreader.val
        else:
            start = 0
            with log_in_place("DistMemMap writing sid_index ", logging.INFO) as updater:
                while start < distreader.sid_count:
                    updater("{0} of {1}".format(start, distreader.sid_count))
                    distdata = distreader[:, start : start + block_size].read(
                        order=order, dtype=dtype, num_threads=num_threads
                    )
                    self.val[:, start : start + distdata.sid_count, :] = distdata.val
                    start += distdata.sid_count

        self.flush()
        if os.path.exists(filename):
            os.remove(filename)
        shutil.move(filename + ".temp", filename)
        logging.debug("Done writing " + filename)
        return DistMemMap(filename)

    def _run_once(self):
        if self._ran_once:
            return
        row_ascii, col_ascii, val, row_property, col_property = self._run_once_inner()
        row = np.array(row_ascii, dtype="str")  # !!!avoid this copy when not needed
        col = np.array(col_ascii, dtype="str")  # !!!avoid this copy when not needed

        DistData.__init__(
            self,
            iid=row,
            sid=col,
            val=val,
            pos=col_property,
            name="np.memmap('{0}')".format(self._filename),
        )


class TestDistMemMap(unittest.TestCase):
    def test1(self):
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        filename2 = "tempdir/tiny.dist.memmap"
        pstutil.create_directory_if_necessary(filename2)
        distreader2 = DistMemMap.empty(
            iid=[["fam0", "iid0"], ["fam0", "iid1"]],
            sid=["snp334", "snp349", "snp921"],
            filename=filename2,
            order="F",
            dtype=np.float64,
        )
        assert isinstance(distreader2.val, np.memmap)
        distreader2.val[:, :, :] = [
            [[0.5, 0.5, 0], [0, 0, 1], [0.5, 0.5, 0]],
            [[0, 1.0, 0], [0, 0.75, 0.25], [0.5, 0.5, 0]],
        ]
        assert np.array_equal(
            distreader2[[1], [1]].read(view_ok=True).val, np.array([[[0, 0.75, 0.25]]])
        )
        distreader2.flush()
        assert isinstance(distreader2.val, np.memmap)
        assert np.array_equal(
            distreader2[[1], [1]].read(view_ok=True).val, np.array([[[0, 0.75, 0.25]]])
        )
        distreader2.flush()

        distreader3 = DistMemMap(filename2)
        assert np.array_equal(
            distreader3[[1], [1]].read(view_ok=True).val, np.array([[[0, 0.75, 0.25]]])
        )
        assert isinstance(distreader3.val, np.memmap)

        logging.info("in TestDistMemMap test1")
        distreader = DistMemMap("../examples/tiny.dist.memmap")
        assert distreader.iid_count == 25
        assert distreader.sid_count == 10
        assert isinstance(distreader.val, np.memmap)

        distdata = distreader.read(view_ok=True)
        assert isinstance(distdata.val, np.memmap)
        os.chdir(old_dir)

    def test2(self):
        from pysnptools.distreader import Bgen

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))

        bgen = Bgen("../examples/example.bgen")
        distmemmap = DistMemMap.write("tempdir/bgentomemmap.dist.memamp", bgen)
        assert DistData.allclose(bgen.read(), distmemmap.read(), equal_nan=True)
        os.chdir(old_dir)

    def test_doctest(self):
        import pysnptools.distreader.distmemmap as mod_mm
        import doctest

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        result = doctest.testmod(mod_mm)
        os.chdir(old_dir)
        assert result.failed == 0, "failed doc test: " + __file__


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestDistMemMap))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    ret = r.run(suites)
    assert ret.wasSuccessful()

    result = doctest.testmod(
        optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
    )
    assert result.failed == 0, "failed doc test: " + __file__



"""Tools for reading and manipulating matrix data along with row and column ids and properties.
"""
from pysnptools.pstreader.pstreader import PstReader
from pysnptools.pstreader.pstdata import PstData
from pysnptools.pstreader.psthdf5 import PstHdf5
from pysnptools.pstreader._oneshot import _OneShot
from pysnptools.pstreader.pstnpz import PstNpz
from pysnptools.pstreader.pstmemmap import PstMemMap
from pysnptools.pstreader._mergerows import _MergeRows
from pysnptools.pstreader._mergecols import _MergeCols




import os
import numpy as np
import logging
import unittest
import doctest
from pathlib import Path
from pysnptools.util.pairs import _Pairs as UtilPairs

from pysnptools.snpreader import SnpReader


class _Pairs(SnpReader):
    """
    Experimental.
    """

    # !!
    # !!!see fastlmm\association\epistasis.py for code that allows ranges of snps to be specified when making pairs
    def __init__(
        self,
        snpreader0,
        snpreader1=None,
        do_standardize=True,
        sid_materialize_limit=1000 * 1000,
        _include_single_times_single=False,
    ):
        # !!! could add option to change snp separator and another to encode chrom, etc in the snp name
        super(_Pairs, self).__init__()
        self._ran_once = False
        self.snpreader0 = snpreader0
        self.snpreader1 = snpreader1
        self.do_standardize = do_standardize
        self.sid_materialize_limit = sid_materialize_limit
        self._include_single_times_single = _include_single_times_single

    def __repr__(self):
        part2 = "" if self.snpreader1 is None else ",{0}".format(self.snpreader1)
        return "{0}({1}{2})".format(self.__class__.__name__, self.snpreader0, part2)

    @property
    def row(self):
        """*same as* :attr:`iid`"""
        if not hasattr(self, "_row"):
            self._row = self.snpreader0.row
            assert self.snpreader1 is None or np.array_equal(
                self._row, self.snpreader1.row
            ), "Expect snpreaders to have the same iids in the same order"
        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`"""
        if not hasattr(self, "_col"):
            assert (
                self.col_count < self.sid_materialize_limit
            ), "{:,} is too many sids to materialize".format(self.col_count)
            if self.snpreader1 is None:
                col_0 = self.snpreader0.col
                col_list = []
                self.index0_list = []  # !!!should be _index0_list, etc
                self.index1_list = []
                for index0 in xrange(self.snpreader0.col_count):
                    # logging.info("index0={0} of {1}".format(index0,self.snpreader0.col_count))# !!!
                    start1 = index0 if self._include_single_times_single else index0 + 1
                    for index1 in xrange(start1, self.snpreader0.col_count):
                        col_list.append("{0},{1}".format(col_0[index0], col_0[index1]))
                        self.index0_list.append(index0)
                        self.index1_list.append(index1)
                self._col = np.array(col_list)
                self.index0_list = np.array(self.index0_list)
                self.index1_list = np.array(self.index1_list)
            else:
                col_0 = self.snpreader0.col
                col_1 = self.snpreader1.col
                assert (
                    len(set(col_0) & set(col_1)) == 0
                ), "Pairs currently requires two snpreaders to have no sids in common"
                col_list = []
                self.index0_list = []
                self.index1_list = []
                for index0 in xrange(self.snpreader0.col_count):
                    # logging.info("index0={0} of {1}".format(index0,self.snpreader0.col_count))# !!!
                    for index1 in xrange(self.snpreader1.col_count):
                        col_list.append("{0},{1}".format(col_0[index0], col_1[index1]))
                        self.index0_list.append(index0)
                        self.index1_list.append(index1)
                self._col = np.array(col_list)
                self.index0_list = np.array(self.index0_list)
                self.index1_list = np.array(self.index1_list)

        assert self.col_count == len(self._col), "real assert"
        return self._col

    @property
    def col_count(self):
        n0 = self.snpreader0.col_count
        if self.snpreader1 is None:
            if self._include_single_times_single:
                return (n0 * n0 + n0) // 2
            else:
                return (n0 * n0 - n0) // 2
        else:
            return n0 * self.snpreader1.col_count

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        if not hasattr(self, "_col_property"):
            self._col_property = np.zeros([self.sid_count, 3], dtype=np.int64)
        return self._col_property

    def copyinputs(self, copier):
        # doesn't need to self._run_once() because only uses original inputs !!!is this true?
        self.snpreader0.copyinputs(copier)
        if self.snpreader1 is not None:
            self.snpreader1.copyinputs(copier)

    def _run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self.col

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self._run_once()
        dtype = np.dtype(dtype)

        iid_count_in = self.iid_count
        sid_count_in = self.sid_count

        if iid_index_or_none is not None:
            iid_count_out = len(iid_index_or_none)
            iid_index = iid_index_or_none
        else:
            iid_count_out = iid_count_in
            iid_index = range(iid_count_in)

        if sid_index_or_none is not None:
            sid_count_out = len(sid_index_or_none)
            sid_index = sid_index_or_none
        else:
            sid_count_out = sid_count_in
            sid_index = range(sid_count_in)

        sid_index_inner_0 = self.index0_list[
            sid_index
        ]  # Find the sid_index of the left snps of interest
        sid_index_inner_1 = self.index1_list[
            sid_index
        ]  # Find the sid_index of the right snps of interest
        if self.snpreader1 is None:
            sid_index_inner_01 = np.unique(
                np.r_[sid_index_inner_0, sid_index_inner_1]
            )  # Index of every snp of interest
            inner_01 = self.snpreader0[iid_index_or_none, sid_index_inner_01].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            val_inner_01 = (
                inner_01.standardize(num_threads=num_threads).val
                if self.do_standardize
                else inner_01.val
            )

            sid_index_inner_01_reverse = {
                v: i for i, v in enumerate(sid_index_inner_01)
            }  # Dictionary of snp_index to position in sid_index_inner_01
            sid_index_inner_0_in_val = np.array(
                [sid_index_inner_01_reverse[i] for i in sid_index_inner_0]
            )  # Replace snp_index0 with column # in val_inner_01
            sid_index_inner_1_in_val = np.array(
                [sid_index_inner_01_reverse[i] for i in sid_index_inner_1]
            )  # Replace snp_index1 with column # in val_inner_01
            val_inner_0 = val_inner_01[
                :, sid_index_inner_0_in_val
            ]  # Extract the vals for the left snps of interest
            val_inner_1 = val_inner_01[
                :, sid_index_inner_1_in_val
            ]  # Extract the vals for the right snps of interest
        else:
            inner_0 = self.snpreader0[iid_index_or_none, sid_index_inner_0].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            inner_1 = self.snpreader1[iid_index_or_none, sid_index_inner_1].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            val_inner_0 = (
                inner_0.standardize(num_threads=num_threads).val
                if self.do_standardize
                else inner_0.val
            )
            val_inner_1 = (
                inner_1.standardize(num_threads=num_threads).val
                if self.do_standardize
                else inner_1.val
            )
        val = (
            val_inner_0 * val_inner_1
        )  # Element multiplication creates the vals for the pairs
        return val


def split_on_sids(snpreader, part_count):
    sid_count = snpreader.sid_count
    start = 0
    for part_index in xrange(1, part_count + 1):
        end = part_index * sid_count // part_count
        yield snpreader[:, start:end]
        start = end


class _Pairs2(SnpReader):
    def __init__(
        self,
        snpreader0,
        snpreader1=None,
        do_standardize=True,
        sid_materialize_limit=1000 * 1000,
        _include_single_times_single=False,
    ):  # !!! could add option to change snp separator and another to encode chrom, etc in the snp name
        super(_Pairs2, self).__init__()
        self._ran_once = False
        self.snpreader0 = snpreader0
        self.snpreader1 = snpreader1
        self.do_standardize = do_standardize
        self.sid_materialize_limit = sid_materialize_limit
        self._include_single_times_single = _include_single_times_single
        self._utilpairs = UtilPairs(
            snpreader0.sid,
            snpreader1.sid if snpreader1 is not None else snpreader0.sid,
            include_singles=_include_single_times_single,
            duplicates_ok=False,
        )

    def __repr__(self):
        part2 = "" if self.snpreader1 is None else ",{0}".format(self.snpreader1)
        return "{0}({1}{2})".format(self.__class__.__name__, self.snpreader0, part2)

    @property
    def row(self):
        """*same as* :attr:`iid`"""
        if not hasattr(self, "_row"):
            self._row = self.snpreader0.row
            assert self.snpreader1 is None or np.array_equal(
                self._row, self.snpreader1.row
            ), "Expect snpreaders to have the same iids in the same order"
        return self._row

    @property
    def col(self):
        """*same as* :attr:`sid`"""
        if not hasattr(self, "_col"):
            assert (
                self.col_count < self.sid_materialize_limit
            ), "{:,} is too many sids to materialize".format(self.col_count)
            snpreader1 = (
                self.snpreader1 if self.snpreader1 is not None else self.snpreader0
            )
            # !!!self.index0_list = self.snpreader0.sid_to_index((sid0 for sid0,sid1 in self._utilpairs[:])) # !!! can we do without these?
            # !!!self.index1_list = snpreader1.sid_to_index(sid1 for sid0,sid1 in self._utilpairs[:])# !!!can we do without these?
            self._col = np.array(list(",".join(pair) for pair in self._utilpairs[:]))
        return self._col

    @property
    def col_count(self):
        return len(self._utilpairs)

    @property
    def col_property(self):
        """*same as* :attr:`pos`"""
        if not hasattr(self, "_col_property"):
            self._col_property = np.zeros([self.sid_count, 3], dtype=np.int64)
        return self._col_property

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs !!!is this true?
        self.snpreader0.copyinputs(copier)
        if self.snpreader1 is not None:
            self.snpreader1.copyinputs(copier)

    def run_once(self):
        if self._ran_once:
            return

        self._ran_once = True
        self.col

    def _read(
        self,
        iid_index_or_none,
        sid_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        self.run_once()

        iid_count_in = self.iid_count
        sid_count_in = self.sid_count

        if iid_index_or_none is not None:
            iid_count_out = len(iid_index_or_none)
            iid_index_out = iid_index_or_none
        else:
            iid_count_out = iid_count_in
            iid_index_out = range(iid_count_in)

        if sid_index_or_none is not None:
            sid_count_out = len(sid_index_or_none)
            sid_index_out = sid_index_or_none
        else:
            sid_count_out = sid_count_in
            sid_index_out = splice(None)  # !!! test this

        pair_array = np.array(
            list(self._utilpairs[sid_index_out])
        )  # !!!make more efficient with npfromiter?
        snpreader1 = self.snpreader1 if self.snpreader1 is not None else self.snpreader0
        sid_index_inner_0 = self.snpreader0.sid_to_index(pair_array[:, 0])
        sid_index_inner_1 = snpreader1.sid_to_index(pair_array[:, 1])

        if self.snpreader1 is None:
            sid_index_inner_01 = np.unique(
                np.r_[sid_index_inner_0, sid_index_inner_1]
            )  # Index of every snp of interest
            inner_01 = self.snpreader0[iid_index_or_none, sid_index_inner_01].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            val_inner_01 = (
                inner_01.standardize().val if self.do_standardize else inner_01.val
            )

            sid_index_inner_01_reverse = {
                v: i for i, v in enumerate(sid_index_inner_01)
            }  # Dictionary of snp_index to position in sid_index_inner_01
            sid_index_inner_0_in_val = np.array(
                [sid_index_inner_01_reverse[i] for i in sid_index_inner_0]
            )  # Replace snp_index0 with column # in val_inner_01
            sid_index_inner_1_in_val = np.array(
                [sid_index_inner_01_reverse[i] for i in sid_index_inner_1]
            )  # Replace snp_index1 with column # in val_inner_01
            val_inner_0 = val_inner_01[
                :, sid_index_inner_0_in_val
            ]  # Extract the vals for the left snps of interest
            val_inner_1 = val_inner_01[
                :, sid_index_inner_1_in_val
            ]  # Extract the vals for the right snps of interest
        else:
            inner_0 = self.snpreader0[iid_index_or_none, sid_index_inner_0].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            inner_1 = self.snpreader1[iid_index_or_none, sid_index_inner_1].read(
                order=order,
                dtype=dtype,
                force_python_only=force_python_only,
                view_ok=True,
                num_threads=num_threads,
            )  # read every val of interest
            val_inner_0 = (
                inner_0.standardize().val if self.do_standardize else inner_0.val
            )
            val_inner_1 = (
                inner_1.standardize().val if self.do_standardize else inner_1.val
            )
        val = (
            val_inner_0 * val_inner_1
        )  # Element multiplication creates the vals for the pairs
        return val


# !!! keep these?


def split_on_sids(snpreader, part_count):
    sid_count = snpreader.sid_count
    start = 0
    for part_index in range(1, part_count + 1):
        end = part_index * sid_count // part_count
        yield snpreader[:, start:end]
        start = end


def epi_reml(
    pair_snps,
    pheno,
    covar=None,
    kernel_snps=None,
    output_dir="results",
    part_count=33,
    runner=None,
    override=False,
):
    from pysnptools.kernelreader import SnpKernel
    from pysnptools.standardizer import Unit
    import datetime
    from fastlmm.association import single_snp

    part_list = list(split_on_sids(pair_snps, part_count))
    part_pair_count = (part_count * part_count + part_count) / 2
    part_pair_index = -1
    print("part_pair_count={0:,}".format(part_pair_count))
    K0 = SnpKernel(
        kernel_snps or pair_snps, standardizer=Unit()
    ).read()  # Precompute the similarity
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    start_time = datetime.datetime.now()
    for i in range(part_count):
        part_i = part_list[i]
        for j in range(i, part_count):
            part_pair_index += 1
            pairs = _Pairs2(part_i) if i == j else _Pairs2(part_i, part_list[j])
            print(
                "Looking at pair {0},{1} which is {2} of {3}".format(
                    i, j, part_pair_index, part_pair_count
                )
            )
            output_file = "{0}/result.{1}.{2}.tsv".format(
                output_dir, part_pair_index, part_pair_count
            )
            if override or not os.path.exists(output_file):
                result_df_ij = single_snp(
                    pairs,
                    K0=K0,
                    pheno=pheno,
                    covar=covar,
                    leave_out_one_chrom=False,
                    count_A1=True,
                    runner=runner,
                )
                result_df_ij.to_csv(output_file, sep="\t", index=False)
                print(result_df_ij[:1])
                time_so_far = datetime.datetime.now() - start_time
                total_time_estimate = (
                    time_so_far * part_pair_count / (part_pair_index + 1)
                )
                print(total_time_estimate)


class TestPairs(unittest.TestCase):
    def test_run1(self):
        from pysnptools.snpreader import Bed

        root = Path(r"D:\OneDrive\programs\epireml")  # !!! make this work without this
        runner = None
        # runner = LocalMultiProc(multiprocessing.cpu_count(),just_one_process=False)

        bed_original = Bed(str(root / "syndata.bed"), count_A1=False)
        pheno = str(root / "pheno.txt")
        covar = str(root / "cov.txt")  # set to None, if none
        output_dir = r"m:/deldir/results.run1"
        part_count = 2

        epi_reml(
            bed_original[:, -20:],
            pheno,
            covar=covar,
            output_dir=output_dir,
            part_count=part_count,
            runner=runner,
            override=True,
        )
        # !!! check answer?

    def test_run2(self):
        from pysnptools.snpreader import Bed

        root = Path(r"D:\OneDrive\programs\epireml")  # !!! make this work without this
        runner = None
        # runner = LocalMultiProc(multiprocessing.cpu_count(),just_one_process=False)
        bed_original = Bed(
            str(root / "syndata.bed"), count_A1=False
        )  # Read only the first 10 SNPs
        pheno = str(root / "pheno.txt")
        covar = str(root / "cov.txt")
        output_dir = r"m:/deldir/results.run2"
        part_count = 1

        epi_reml(
            bed_original[:, :20],
            pheno,
            kernel_snps=bed_original,
            covar=covar,
            output_dir=output_dir,
            part_count=part_count,
            runner=runner,
            override=True,
        )

        # !!! check answer? against M:\deldir\refresults.run1 and 2


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPairs))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    TestPairs().test_run1()
    TestPairs().test_run2()

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    r.run(suites)

    result = doctest.testmod()
    assert result.failed == 0, "failed doc test: " + __file__

    print("done")

    from pysnptools.snpreader import Pheno, Bed
    import pysnptools.util as pstutil

    if False:
        from pysnptools.snpreader import Bed
        from pysnptools.util import example_file  # Download and return local file name

        bed_file = example_file("doc/ipynb/all.*", "*.bed")
        bed = Bed(bed_file, count_A1=False)
        print(f"{bed.sid_count:,}")
        pairs = _Pairs(bed)
        print(f"{pairs.sid_count:,}")
        subset = pairs[:10, 10000:10010]
        print(subset.read().val.shape)
        print(subset.sid)

    if False:
        data_file = "d:\OneDrive\programs\epiCornell\syndata.bed"
        if False:
            from pysnptools.snpreader import SnpData
            import numpy as np

            bed1 = Bed("../../tests/datasets/synth/all")
            print(bed1.iid_count, bed1.sid_count, bed1.iid_count * bed1.sid_count)
            # goal 1500 individuals x 27000 SNP
            snpdata1 = bed1.read()
            iid = bed1.iid
            sid = ["sid{0}".format(i) for i in xrange(27000)]
            val = np.tile(snpdata1.val, (3, 6))[:, :27000].copy()
            # snpdata = Pheno('pysnptools/examples/toydata.phe').read()         # Read data from Pheno format
            snpdata2 = SnpData(iid, sid, val)
            print(
                snpdata2.iid_count,
                snpdata2.sid_count,
                snpdata2.iid_count * snpdata2.sid_count,
            )
            Bed.write(snpdata2, data_file, count_A1=False)

        synbed = Bed(data_file)
        print(synbed.iid_count, synbed.sid_count, synbed.iid_count * synbed.sid_count)

        part_count = 1000
        part_list = list(split_on_sids(synbed, part_count))

        pairs00 = _Pairs(part_list[0])
        from fastlmm.association import single_snp

        pheno_fn = r"d:\OneDrive\programs\epiCornell\pheno.txt"
        cov_fn = r"d:\OneDrive\programs\epiCornell\cov.txt"
        results_df = single_snp(
            pairs00,
            K0=synbed,
            pheno=pheno_fn,
            covar=cov_fn,
            leave_out_one_chrom=False,
            count_A1=True,
        )

        if False:
            for i, synbed_part_i in enumerate(synbed_part_list):
                for j, synbed_part_j in enumerate(synbed_part_list):
                    if j < i:
                        continue  # not break
                    print("Looking at pair {0},{1}".format(i, j))
                    pairs = (
                        _Pairs(synbed_part_i)
                        if i == j
                        else _Pairs(synbed_part_i, synbed_part_j)
                    )
                    # print(pairs.iid)
                    print("{:,}".format(pairs.sid_count))
                    # print(pairs.sid)
                    # print(pairs.pos)
                    # print(pairs.row_property)
                    snpdata = pairs.read()  #
                    # print(snpdata.val)

        import datetime
        from pysnptools.kernelreader import SnpKernel
        from pysnptools.standardizer import Unit
        from pysnptools.util.mapreduce1.runner import LocalMultiProc
        from pysnptools.util.mapreduce1 import map_reduce

        # runner=None
        runner = LocalMultiProc(1, just_one_process=False)

        part_pair_count = (part_count * part_count + part_count) // 2
        part_pair_index = -1
        print("part_pair_count={0:,}".format(part_pair_count))

        K0 = SnpKernel(synbed, standardizer=Unit()).read()  # Precompute the similarity

        start_time = datetime.datetime.now()
        for i, part_i in enumerate(part_list):

            def mapper1(j):
                # from fastlmm.association import single_snp
                # from pysnptools.snpreader import Pairs
                # print('Z')
                # part_j = part_list[j]
                # print('A')
                print(
                    "Looking at pair {0},{1} which is {2} of {3}".format(
                        i, j, part_pair_index + j + 1, part_pair_count
                    )
                )
                # pairs = Pairs(part_i) if i==j else Pairs(part_i,part_j)
                # result_df_ij = single_snp(pairs, K0=K0, pheno=pheno_fn, covar=cov_fn, leave_out_one_chrom=False, count_A1=True)
                # print(result_df_ij[:1])
                # return result_df_ij

            result_df_i = map_reduce(
                range(i, part_count),
                mapper=mapper1,
                reducer=lambda result_j_list: pd.concat(result_j_list),
                runner=runner,
                name="js",
            )
            part_pair_index += part_count - i
            time_so_far = datetime.datetime.now() - start_time
            total_time_estimate = time_so_far * part_pair_count / (part_pair_index + 1)
            print(total_time_estimate)



import numpy as np
import subprocess, sys, os.path
from itertools import *
import pandas as pd
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.snpreader import SnpData
import numpy as np
import warnings
from pysnptools.pstreader import _OneShot


class Ped(_OneShot,SnpReader):
    '''
    A :class:`.SnpReader` for reading Ped-formated (and the related Map-formated) files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    This format is described `here <http://zzz.bwh.harvard.edu/plink/data.shtml#ped>`__ and looks like::

         FAM001  1  0 0  1  2  A A  G G  A C 
         FAM001  2  0 0  1  2  A A  A G  0 0 
         ...

    the direction of the encoding from allele pair to 0,1,2 is arbitrary. That is, if the alleles are "A" and "G", then "G G" could be 0 and "A A" could be 2 or visa versa. The pair "A G" will always be 1.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The Ped file to read.
                     * **missing** (*string*) -- The value in the file that represents missing data.


        :Example:

        >>> from pysnptools.snpreader import Ped
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> ped_file = example_file('pysnptools/examples/toydata.ped')
        >>> data_on_disk = Ped(ped_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 10000)

    **Methods beyond** :class:`.SnpReader`
    '''

    def __init__(self, filename, missing = '0'):
        '''
            filename    : string of the filename of the ped file
            missing         : string indicating a missing genotype (default '0')
        '''
        super(Ped, self).__init__()
        self.filename = SnpReader._name_of_other_file(filename,remove_suffix="ped", add_suffix="ped")
        self.missing = missing

    def _read_pstdata(self):
        col, col_property = SnpReader._read_map_or_bim(self.filename,remove_suffix="ped", add_suffix="map")
        ped = np.loadtxt(self.filename, dtype='str', comments=None)
        ped = ped.reshape(-1,ped.shape[-1]) #Turns 1-d row into 2-d
        row = ped[:,0:2]
        snpsstr = ped[:,6::]
        inan=snpsstr==self.missing
        snps = np.zeros((snpsstr.shape[0],snpsstr.shape[1]//2))
        for i in range(snpsstr.shape[1]//2):
            snps[inan[:,2*i],i]=np.nan
            vals=snpsstr[~inan[:,2*i],2*i:2*(i+1)]
            if vals.shape[0] > 0:
                snps[~inan[:,2*i],i]+=(vals==vals[0,0]).sum(1)
        snpdata = SnpData(iid=row,sid=col,pos=col_property,val=snps)
        return snpdata

    def copyinputs(self, copier):
        # doesn't need to self.run_once() because only uses original inputs
        copier.input(self.filename)
        copier.input(SnpReader._name_of_other_file(self.filename,remove_suffix="ped", add_suffix="map"))


    @staticmethod
    def write(filename, snpdata):
        """Writes a :class:`SnpData` to Ped format. The values must be 0,1,2. The direction of the encoding to allele pairs is arbitrary. This means
        that if a SnpData is written in Ped format and then read back, then 0's may become 2's and 2's may become 0's. (1's will stay 1's).
        Returns the :class:`.Ped`

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :rtype: :class:`.Ped`

        >>> from pysnptools.snpreader import Ped, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:,:10].read()  # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.ped")
        >>> Ped.write("tempdir/toydata10.ped",snpdata)            # Write data in Ped format
        Ped('tempdir/toydata10.ped')
        """

        if isinstance(filename,SnpData) and isinstance(snpdata,str): #For backwards compatibility, reverse inputs if necessary
            warnings.warn("write statement should have filename before data to write", DeprecationWarning)
            filename, snpdata = snpdata, filename 

        SnpReader._write_map_or_bim(snpdata, filename, remove_suffix="ped", add_suffix="map")

        # The PED file is a white-space (space or tab) delimited file: the first six columns are mandatory:
        # Family ID
        # Case ID
        # Paternal ID
        # Maternal ID
        # Sex (1=male; 2=female; other=unknown)
        # Phenotype

        pedfile = SnpReader._name_of_other_file(filename, remove_suffix="ped", add_suffix="ped")
        with open(pedfile,"w") as ped_filepointer:
            for iid_index, iid_row in enumerate(snpdata.iid):
                ped_filepointer.write("{0} {1} 0 0 0 0".format(iid_row[0],iid_row[1]))
                row = snpdata.val[iid_index,:]
                for sid_index, val in enumerate(row):
                    if val == 0:
                        s = "A A"
                    elif val == 1:
                        s = "A G"
                    elif val == 2:
                        s = "G G"
                    elif np.isnan(val):
                        s = "0 0"
                    else:
                        raise Exception("Expect values for ped file to be 0,1,2, or NAN. Instead, saw '{0}'".format(val))
                    ped_filepointer.write("\t"+s)
                ped_filepointer.write("\n")
        return Ped(filename)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from itertools import chain, islice
from more_itertools import unique_everseen
import doctest
import unittest
import logging
import numbers


class _Pairs(object):
    """!!!description of class"""

    def __init__(
        self, list0, list1, include_singles, duplicates_ok=True
    ):  # !!! rename list to iterable?
        self.include_singles = include_singles
        self.duplicates_ok = duplicates_ok
        self.list0 = list(unique_everseen(list0))
        self.list1 = list(unique_everseen(list1))
        if not duplicates_ok:
            assert len(list0) == len(self.list0) and len(list1) == len(
                self.list1
            ), "Expect lists to have no duplicates"

        dict0 = {k: v for v, k in enumerate(self.list0)}
        dict1 = {k: v for v, k in enumerate(self.list1)}
        self._common = sorted(dict0.keys() & self.list1, key=lambda k: dict0[k])
        self._only0 = sorted(dict0.keys() - self._common, key=lambda k: dict0[k])
        self._only1 = sorted(dict1.keys() - self._common, key=lambda k: dict1[k])

        # How many pairs?
        self._count0 = len(self._only0) * len(self.list1)
        self._count1 = self._count1_fun(
            len(self._common), len(self._common), len(self._only1), include_singles
        )
        self._count = self._count0 + self._count1

    def __len__(self):
        return self._count

    def __getitem__(self, indexer, speed="fast"):
        if isinstance(indexer, slice):
            assert indexer.step is None, "step is not supported"
            start, stop = indexer.start, indexer.stop
            if start is None:
                start = 0
            if stop is None:
                stop = self._count
            if start < 0:
                start = max(0, self._count + start)
            if stop < 0:
                stop = self._count + stop
            for pair in self._pair_sequence(start, stop, speed=speed):
                yield pair
        elif isinstance(indexer, numbers.Integral):
            start = indexer
            if start < 0:
                start = self._count + start
            assert (
                0 <= start and start < self._count
            ), "Expect index between 0 (inclusive) and count of pairs (exclusive)"
            stop = start + 1
            for pair in self._pair_sequence(start, stop, speed=speed):
                yield pair
        elif isinstance(
            indexer, np.ndarray
        ):  # !!! what about other arrays? iterator? negative elements?
            # !!! need test for this
            iterator = None
            previous = None
            for i in indexer:
                if iterator is None or i != previous + 1:
                    iterator = self._pair_sequence(i, None, speed=speed)
                yield next(iterator)
                previous = i
        else:
            raise Exception("Don't understand indexer '{0}'".format(indexer))

    def _pair_sequence(self, start=0, stop=None, speed="fast"):
        start = int(start)
        if speed == "fast":
            stop = min(stop, self._count) if stop is not None else self._count
            return islice(self._pair_sequence_inner(start=start), max(stop - start, 0))
        if speed == "medium":
            stop = stop if stop is not None else self._count
            return islice(
                self._pair_sequence_inner_medium(start=start), max(stop - start, 0)
            )
        else:
            assert speed == "slow", "Don't know speed '{0}'".format(speed)
            return islice(self._pair_sequence_inner_slow(), start, stop)

    def _pair_sequence_inner(self, start=0):
        if start >= self._count0:
            start -= self._count0
        else:
            row_start0 = start // len(self.list1)
            start -= row_start0 * len(self.list1)
            assert 0 <= start and start <= len(self.list1), "real assert"
            for v0 in self._only0[row_start0:]:
                for v1 in islice(chain(self._only1, self._common), start, None):
                    yield v0, v1
                start = 0

        row_start1 = self._row_start1_fun(
            start, len(self._common), len(self._only1), self.include_singles
        )
        start -= self._count1_fun(
            row_start1, len(self._common), len(self._only1), self.include_singles
        )
        assert 0 <= start and start <= len(self.list1), "real assert"
        for common_index in range(row_start1, len(self._common)):
            v0 = self._common[common_index]
            common_start = common_index if self.include_singles else common_index + 1
            assert start < len(self.list1) - common_start, "real assert"
            for v1 in islice(
                chain(self._only1, self._common[common_start:]), start, None
            ):
                yield v0, v1
            start = 0

    def _pair_sequence_inner_medium(self, start=0):
        assert 0 <= start, "real assert"
        for v0 in self._only0:
            if start > len(self.list1):
                start -= len(self.list1)
            else:
                for v1 in islice(chain(self._only1, self._common), start, None):
                    yield v0, v1
                start = 0

        assert 0 <= start, "real assert"
        for index in range(len(self._common)):
            v0 = self._common[index]
            startx = index if self.include_singles else index + 1
            if start > len(self.list1) - startx:
                start -= len(self.list1) - startx
            else:
                for v1 in islice(
                    chain(self._only1, self._common[startx:]), start, None
                ):
                    yield v0, v1
                start = 0

    def _pair_sequence_inner_slow(self):
        for v0 in self._only0:
            for v1 in chain(self._only1, self._common):
                yield v0, v1

        for index in range(len(self._common)):
            v0 = self._common[index]
            startx = index if self.include_singles else index + 1
            for v1 in chain(self._only1, self._common[startx:]):
                yield v0, v1

    @staticmethod
    def _count1_fun(row_start, len_common, len_only1, include_singles):
        a2 = -1
        b2 = len_only1 * 2 + len_common * 2 + (1 if include_singles else -1)
        count1 = (a2 * row_start * row_start + b2 * row_start) // 2
        return count1

    @staticmethod
    def _row_start1_fun(count1, len_common, len_only1, include_singles):
        a2 = -1
        b2 = len_only1 * 2 + len_common * 2 + (1 if include_singles else -1)
        c = -count1
        row_start = int(
            (b2 - (b2 * b2 - 8 * a2 * c) ** 0.5) / 2.0
        )  # !!! will this ever be off by one because of numerical problems?
        if row_start < 0:
            row_start = int((b2 + (b2 * b2 - 8 * a2 * c) ** 0.5) / 2.0)
        return row_start


class TestPairs(unittest.TestCase):
    def test_count1_fun(self):
        for len_common in range(5):
            for len_only1 in range(5):
                for include_singles in [True, False]:
                    list0 = ["common{0}".format(i) for i in range(len_common)]
                    list1 = list0 + ["only1_{0}".format(i) for i in range(len_only1)]
                    pairs = _Pairs(list0, list1, include_singles)
                    pair_list = list(pairs._pair_sequence(0, speed="slow"))
                    for row_start in range(len_common + 1):
                        count = _Pairs._count1_fun(
                            row_start, len_common, len_only1, include_singles
                        )
                        count_ref = len(
                            [pair for pair in pair_list if pair[0] in list0[:row_start]]
                        )
                        assert (
                            count == count_ref
                        ), "_count1_fun isn't giving the right answer"

    def test_row_start_fun(self):
        for len_common in range(5):
            for len_only1 in range(5):
                for include_singles in [True, False]:
                    list0 = ["common{0}".format(i) for i in range(len_common)]
                    list1 = list0 + ["only1_{0}".format(i) for i in range(len_only1)]
                    pairs = _Pairs(list0, list1, include_singles)
                    pair_list = list(pairs._pair_sequence(0, speed="slow"))
                    for start in range(len(pairs) + 1):
                        row_start = pairs._row_start1_fun(
                            start, len_common, len_only1, include_singles
                        )
                        assert (
                            start == len(pair_list)
                            or pair_list[start][0] == list0[row_start]
                        ), "_row_start1_fun isn't giving the right answer"

    def test_index_and_count1_functions(self):
        for len_commonq in range(0, 5):
            for len_only1q in range(0, 5):
                for include_singlesq in [True, False]:
                    for row_startq in range(len_commonq + 1):
                        count1q = _Pairs._count1_fun(
                            row_startq, len_commonq, len_only1q, include_singlesq
                        )
                        row_start2 = _Pairs._row_start1_fun(
                            count1q, len_commonq, len_only1q, include_singlesq
                        )
                        count2 = _Pairs._count1_fun(
                            row_start2, len_commonq, len_only1q, include_singlesq
                        )
                        assert count1q == count2

    def test_medium_fast(self):
        for len_common in range(2):
            common = ["common{0}".format(i) for i in range(len_common)]
            for len_only0 in range(2):
                list0 = common + ["only0_{0}".format(i) for i in range(len_only0)]
                for len_only1 in range(2):
                    list1 = common + ["only1_{0}".format(i) for i in range(len_only1)]
                    for include_singles in [True, False]:
                        pairs = _Pairs(
                            list0, list1, include_singles, duplicates_ok=True
                        )
                        for goal_see in range(len(pairs) + 1):
                            for start in range(len(pairs) + 1):
                                logging.info(
                                    (
                                        len_common,
                                        len_only0,
                                        len_only1,
                                        include_singles,
                                        start,
                                        start + goal_see,
                                    )
                                )
                                slow = np.array(
                                    list(
                                        pairs._pair_sequence(
                                            start, start + goal_see, speed="slow"
                                        )
                                    )
                                )
                                medium = np.array(
                                    list(
                                        pairs._pair_sequence(
                                            start, start + goal_see, speed="medium"
                                        )
                                    )
                                )
                                assert np.array_equal(
                                    slow, medium
                                ), "Expect slow and medium to give the same answers"
                                fast = np.array(list(pairs[start : start + goal_see]))
                                assert np.array_equal(
                                    medium, fast
                                ), "Expect medium and fast to give the same answers"

    def test_big(self):
        size = 500 * 1000
        seed = 0
        np.random.seed(seed)
        list0 = np.random.randint(size * 10, size=size)
        list1 = np.random.randint(size * 10, size=size)
        for include_singles in [True, False]:
            pairs = _Pairs(list0, list1, include_singles, duplicates_ok=True)
            for start in [0, len(pairs) // 5, len(pairs) - 1, len(pairs)]:
                logging.info(("test_big", include_singles, start))
                medium = np.array(
                    list(pairs._pair_sequence(start, start + 10, speed="medium"))
                )
                fast = np.array(list(pairs[start : start + 10]))
                assert np.array_equal(
                    medium, fast
                ), "Expect medium and fast to give the same answers"

    # !!! make doc string example with first and last names
    def test_slice(self):
        for len_common in range(3):
            common = ["common{0}".format(i) for i in range(len_common)]
            for len_only0 in range(3):
                list0 = common + ["only0_{0}".format(i) for i in range(len_only0)]
                for len_only1 in range(3):
                    list1 = common + ["only1_{0}".format(i) for i in range(len_only1)]
                    for include_singles in [True, False]:
                        pairs = _Pairs(
                            list0, list1, include_singles, duplicates_ok=True
                        )
                        pair_list = np.array(
                            list(pairs._pair_sequence(0, None, speed="slow")), dtype="U"
                        )
                        for slicer, start, stop in [
                            (-1, len(pairs) - 1, len(pairs)),  # [-1]
                            (slice(None, None), 0, len(pairs)),  # [:]
                            (slice(None, 2), 0, 2),  # [:2]
                            (2, 2, 3),  # [2]
                            (slice(2, None), 2, None),  # [2:]
                            (slice(1, 3), 1, 3),  # [1:3]
                            (slice(3, 3), 3, 3),  # [3:3]
                            (slice(-3, -2), len(pairs) - 3, len(pairs) - 2),  # [-3:-2]
                            (slice(-3, None), len(pairs) - 3, None),  # [-3:],
                            (slice(None, -3), None, len(pairs) - 3),  # [:-3],
                            (slice(1000, None), 1000, None),  # [1000:]
                            (slice(-3, None), len(pairs) - 100000, None),  # [-100000:]
                            (slice(4, 1), 4, 1),  # [4:1],
                            (slice(10, 1000), 10, 1000),  # [10,1000]
                        ]:
                            logging.info(
                                (
                                    "test_slice",
                                    slicer,
                                    start,
                                    stop,
                                    len_common,
                                    len_only0,
                                    len_only1,
                                    include_singles,
                                )
                            )
                            try:
                                slow = pair_list[slicer]
                                slow_fail = False
                            except Exception:
                                slow_fail = True
                            try:
                                fast = np.array(list(pairs[slicer]), dtype="U")
                                fast_fail = False
                            except Exception:
                                fast_fail = True
                            if slow_fail:
                                assert (
                                    fast_fail
                                ), "expect slow and fast to fail on the same inputs"
                            else:
                                if fast_fail or not np.array_equal(
                                    slow.reshape(-1, 2), fast.reshape(-1, 2)
                                ):  # !!!:
                                    fast = np.array(list(pairs[slicer]))
                                assert (
                                    not fast_fail
                                ), "expect slow and fast to fail on the same inputs"
                                assert np.array_equal(
                                    slow.reshape(-1, 2), fast.reshape(-1, 2)
                                ), "Expect slow and fast to give the same answers"


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestPairs))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    # !!! this is work on an example for docstring, but a bit weird because names are ordered and these arent
    pairs = _Pairs(
        ["Madison", "Jill", "Kelsey"],
        ["Smith", "Madison", "Kelsey"],
        include_singles=True,
    )  # !!! is 'include_singles' the best name?
    for first, last in pairs[:]:
        print(first, last)

    pairs = _Pairs(range(0, 2000), range(1000, 3000), include_singles=True)
    print(len(pairs))
    for pair in pairs[1999999:2000004]:
        print(pair)
    print(list(pairs[-5:]))

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=True)
    r.run(suites)

    result = doctest.testmod()
    assert result.failed == 0, "failed doc test: " + __file__

    print("done")



import numpy as np
import logging
import warnings
from pysnptools.standardizer import Standardizer


class Beta(Standardizer):
    """
    A :class:`.Standardizer` to beta standardize SNP data.

    See :class:`.Standardizer` for more information about standardization.

    **Constructor:**
        :Parameters: * **a** (*float*) -- The *a* parameter of the beta distribution
                     * **b** (*float*) -- The *b* parameter of the beta distribution

    >>> from pysnptools.standardizer import Beta
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read().standardize(Beta(1,25))
    >>> print('{0:.6f}'.format(snpdata1.val[0,0]))
    0.680802
    """

    def __init__(self, a, b):
        super(Beta, self).__init__()
        self.a = a
        self.b = b

    def __repr__(self):
        return "{0}(a={1},b={2})".format(self.__class__.__name__, self.a, self.b)

    def standardize(
        self,
        snpdata,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):  # !!!later why is the 2nd argument called 'snpdata' here, but 'snps' in unit.py?
        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )

        if hasattr(snpdata, "val"):
            val = snpdata.val
        else:
            warnings.warn(
                "standardizing an nparray instead of a SnpData is deprecated",
                DeprecationWarning,
            )
            val = snpdata

        stats = self._standardize_unit_and_beta(
            val,
            is_beta=True,
            a=self.a,
            b=self.b,
            apply_in_place=True,
            use_stats=False,
            stats=None,
            num_threads=num_threads,
            force_python_only=force_python_only,
        )
        if return_trained:
            from pysnptools.standardizer import BetaTrained

            assert hasattr(
                snpdata, "val"
            ), "return_trained=True must be used with SnpData"
            return snpdata, BetaTrained(self.a, self.b, snpdata.sid, stats)
        else:
            return snpdata

    def _merge_trained(self, trained_list):
        from pysnptools.standardizer import BetaTrained

        sid = np.concatenate([trained.sid for trained in trained_list])
        stats = np.concatenate([trained.stats for trained in trained_list])
        a_set = set([trained.a for trained in trained_list])
        b_set = set([trained.b for trained in trained_list])
        assert len(a_set) <= 1, "Expect all BetaTrained's to have the same 'a'"
        assert len(b_set) <= 1, "Expect all BetaTrained's to have the same 'b'"
        a = list(a_set)[0] if a_set else None
        b = list(b_set)[0] if b_set else None
        return BetaTrained(a, b, sid, stats)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from unittest.mock import patch
import os, sys
import logging
from pysnptools.util.mapreduce1.runner import Runner, _JustCheckExists, _run_one_task
from pysnptools.util import create_directory_if_necessary
import cloudpickle as pickle


class LocalInParts(Runner):
    """
    A :class:`.Runner` that runs one piece of a :func:`.map_reduce` job locally. Partial results are saved to disk.
    Clustering runners and :class:`LocalMultiProc` use this runner internally.

    **Constructor:**
        :Parameters: * **taskindex** (*number*) -- Which piece of work to run. When 0 to **taskcount**-1, does map work. When **taskcount**, does the reduce work.
                     * **taskcount** (*number*) -- The number of pieces into which to divide the work.
                     * **mkl_num_threads** (*number*) -- (default None) Limit on the number threads used by the NumPy MKL library.
                     * **weights** (*array of integers*) -- (default None) If given, tells the relative amount of work assigned to
                              each task. The length of the array must be **taskcount**. If not given, all tasks are assigned the
                              same amount of work.
                     * **environ** (*dictionary of variables and values*) -- (default None). Temporarily assigns environment variables.
                              both variables and values should be strings.
                     * **result_file** (*string*) -- (default None) Where to pickle the final results. If no file is given, the final results are returned, but not saved to a file.
                     * **result_dir** (*string*) -- (default None) The directory for any result_file. Defaults to the current working directory.
                     * **temp_dir** (*string*) -- (default None) The directory for partial results. Defaults to the **result_dir**/.working_directory.{map_reduce's Name}.
                     * **logging_handler** (*stream*) --  (default stdout) Where to output logging messages.

        :Example:

        >>> from pysnptools.util.mapreduce1 import map_reduce
        >>> from pysnptools.util.mapreduce1.runner import LocalInParts
        >>> def holder1(n,runner):
        ...     def mapper1(x):
        ...         return x*x
        ...     def reducer1(sequence):
        ...        return sum(sequence)
        ...     return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        >>> holder1(100,LocalInParts(0,4)) #Run part 0 of 4 and save partial results to disk as '0.4.p'.
        >>> holder1(100,LocalInParts(1,4)) #Run part 1 of 4 and save partial results to disk as '1.4.p'.
        >>> holder1(100,LocalInParts(2,4)) #Run part 2 of 4 and save partial results to disk as '2.4.p'.
        >>> holder1(100,LocalInParts(3,4)) #Run part 3 of 4 and save partial results to disk as '3.4.p'.
        >>> holder1(100,LocalInParts(4,4)) #Read the all partial results and then apply the reduce function and return the result.
        328350

    """

    def __init__(
        self,
        taskindex,
        taskcount,
        mkl_num_threads=None,
        weights=None,
        environ=None,
        result_file=None,
        run_dir=".",
        temp_dir=None,
        logging_handler=logging.StreamHandler(sys.stdout),
    ):
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)
        logger.addHandler(logging_handler)

        self.temp_dir = temp_dir
        self.run_dir = run_dir
        self.result_file = os.path.join(run_dir, result_file) if result_file else None
        self.taskindex = taskindex
        self.taskcount = taskcount
        self.weights = weights
        self.environ = environ
        self.mkl_num_threads = mkl_num_threads

    def run(self, distributable):
        if self.temp_dir is not None:
            tempdir = self.temp_dir
        else:
            tempdir = os.path.join(self.run_dir, distributable.tempdirectory)
        tempdir = os.path.realpath(tempdir)

        with patch.dict(
            "os.environ",
            {"MKL_NUM_THREADS": str(self.mkl_num_threads)}
            if self.mkl_num_threads is not None
            else {},
        ) as _:
            if self.taskindex != self.taskcount:
                _JustCheckExists().input(distributable)
                return _run_one_task(
                    distributable,
                    self.taskindex,
                    self.taskcount,
                    tempdir,
                    weights=self.weights,
                    environ=self.environ,
                )
            else:
                result = _run_one_task(
                    distributable,
                    self.taskindex,
                    self.taskcount,
                    tempdir,
                    weights=self.weights,
                    environ=self.environ,
                )
                if self.result_file is not None:
                    create_directory_if_necessary(self.result_file)
                    with open(self.result_file, mode="wb") as f:
                        pickle.dump(result, f)

                return result


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.kernelreader import KernelReader
from pysnptools.pstreader._subset import _PstSubset

class _KernelSubset(KernelReader,_PstSubset):
    def __init__(self, *args, **kwargs):
        super(_KernelSubset, self).__init__(*args, **kwargs)

